/*
 *  linux/fs/nfs/nfs4xdr.c
 *
 *  xdr functions to encode/decode nfs version 4 arguments and results.
 *
 *  by Andy Adamson <andros@citi.umich.edu>
 *  Copyright (C) 1999 The Regents of the University of Michigan
 *
 *  $Id: nfs4xdr.c,v 1.11 2000/08/09 23:08:07 rees Exp $
 */


#define NFS_NEED_NFS4_XDR_TYPES

#ifdef __linux__
#include <linux/param.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/utsname.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/in.h>
#include <linux/pagemap.h>
#include <linux/proc_fs.h>
#include <linux/kdev_t.h>
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/msg_prot.h>
#include <linux/nfs.h>
#include <linux/nfs3.h>
#include <linux/nfs4_prot.h>
#include <linux/nfs_fs.h>
#endif

#ifdef __OpenBSD__
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/kernel.h>
#include <sys/mount.h>
#include <sys/dirent.h>
#include <sys/buf.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/systm.h>

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>

#include <nfs/rpcv2.h>
#include <nfs/nfsproto.h>
#include <nfs/nfsnode.h>
#include <nfs/nfs.h>
#include <nfs/nfsmount.h>
#include <nfs/xdr_subs.h>
#include <nfs/nfsm_subs.h>
#include <nfs/nfsdiskless.h>
#include <nfs/nqnfs.h>
#include <nfs/nfs_var.h>
#include <nfs/nfs4_prot.h>
#endif


#define NFSDBG_FACILITY		NFSDBG_XDR


/*
extern int			nfs_stat_to_errno(int);
*/

/*
 * Declare the space requirements for NFS arguments and replies as
 * number of 32bit-words
 */
#define NFS_enc_void_sz			0
#define NFS_dec_void_sz			0
/* initial size. real size set by nfs4_set_bsize()*/
#define NFS_compoundargs_sz     1
#define NFS_compoundres_sz      1



void
print_lockowner(nfs_lockowner4 *lowner)
{
	nfsv4_printk(lock,"lockowner clientid: [0x%lx , %08lx]\n",
		(unsigned long)(lowner->clientid >> 32),
		(unsigned long)(lowner->clientid & 0xFFFFFFFFUL));
	nfsv4_printk(lock,"lockowner owner: 0x%x\n",*(pid_t *)lowner->owner.owner_val);
}

void
print_utf8string(utf8string *str)
{
char *ptr = str->utf8string_val;
int i;
	if(str->utf8string_val == NULL){
		nfsv4_printk(utf8,"NFSV4 print_utf8string, NULL val\n");
		return;
	}
	for(i=0;i<str->utf8string_len;i++){
		nfsv4_printk(utf8,"%c",*ptr);
		ptr++;
	}
	nfsv4_printk(utf8,"\n");
}

/*
 * Common NFS XDR functions as inlines
 */

static inline u32 *
xdr_encode_64(u32 *p, u64 val)
{
	nfsv4_printk(xdr,"NFSDV4 xdr_encode_64 val: %ld\n",(long int)val);
	*p++ = htonl(val >> 32);
	*p++ = htonl(val & 0xffffffff);
	return p;
}

static inline u32 *
xdr_decode_64(u32 *p, u64 *valp)
{
	*valp  = ((u64) ntohl(*p++)) << 32;
	*valp |= ntohl(*p++);
	return p;
}
static inline u32 *
xdr_decode_64_inplace(u32 *p)
{
u64 temp = 0;
u32 *rem;
	rem = p;
	p = xdr_decode_64(p,&temp);
	*(u64 *)rem = temp;
	return p;
}

static inline u32 *
xdr_encode_utf8string(u32 *p, utf8string  *utf)
{
int quadlen;

	quadlen = XDR_QUADLEN(utf->utf8string_len);
	p[quadlen] = 0;
	*p++=htonl(utf->utf8string_len);	
	memcpy(p,utf->utf8string_val,utf->utf8string_len);
	return p + quadlen;
}

static u32 *
xdr_decode_utf8string(u32 *p, utf8string *utf)
{
	utf->utf8string_len = ntohl(*p++);
	utf->utf8string_val = (char *)p;
	return p + XDR_QUADLEN(utf->utf8string_len);
}


static inline u32 *
xdr_encode_fh4(u32 *p, nfs_fh4 *fhp)
{
	nfsv4_printk(fh4,"NFSV4: xdr_encode_fh4: FILE HANDLE:\n");
	print_nfs_fh4(fhp);
	if(fhp->nfs_fh4_len <= NFS4_FHSIZE){
		*p++ = htonl(fhp->nfs_fh4_len);
		memcpy(p, fhp->nfs_fh4_val, fhp->nfs_fh4_len);
		return p + XDR_QUADLEN(fhp->nfs_fh4_len);
	}
	else 
		return NULL;
}

static inline u32 *
xdr_decode_fh4(u32 *p, nfs_fh4 *fhp)
{
	fhp->nfs_fh4_len = ntohl(*p++);
	if (fhp->nfs_fh4_len <= NFS4_FHSIZE) {
		fhp->nfs_fh4_val= kmalloc(fhp->nfs_fh4_len,GFP_KERNEL);
		memset(fhp->nfs_fh4_val,0,fhp->nfs_fh4_len);
		memcpy(fhp->nfs_fh4_val, p, fhp->nfs_fh4_len);
		print_nfs_fh4(fhp);
		return p + XDR_QUADLEN(fhp->nfs_fh4_len);
	} else 
		memset((u8 *)fhp, 0, sizeof(*fhp));
	return NULL;
}


static inline u32 *
xdr_encode_bitmap4(u32 *p, bitmap4  *bm4)
{
int i;
uint32_t *ptr=bm4->BVAL;

	nfsv4_printk(xdr,"NFSV4 xdr_encode_bitmap4 len %d\n",bm4->BLEN);
	*p++ =  htonl(bm4->BLEN);
	for(i=0;i<bm4->BLEN;i++){
		*p++ = htonl(*ptr++);
	}
	return p;
}

static inline u32 *
xdr_encode_nfs_fh4(u32 *p, nfs_fh4 *fh)
{
	*p++ = htonl(fh->nfs_fh4_len);
	memcpy(p, fh->nfs_fh4_val, fh->nfs_fh4_len);
	return p + XDR_QUADLEN(fh->nfs_fh4_len);
}


static inline u32 *
xdr_decode_nfs_fh4_inplace(u32 *p)
{
int len;
	len = *p = ntohl(*p++);
	if(len <= NFS4_FHSIZE)
		return p + XDR_QUADLEN(len);
	else
	return NULL;
}
static inline u32 *
xdr_decode_nfs_fh4(u32 *p, nfs_fh4 *fh)
{
	/*
	 * Zero all nonused bytes
	 */
	if ((fh->nfs_fh4_len = ntohl(*p++)) <= NFS4_FHSIZE) {
		memcpy(fh->nfs_fh4_val, p, fh->nfs_fh4_len);
		return p + XDR_QUADLEN(fh->nfs_fh4_len);
	} else
		memset((u8 *)fh, 0, sizeof(*fh));
	return NULL;
}

/*
 * Encode/decode time.
 * Since the VFS doesn't care for fractional times, we ignore the
 * nanosecond field.
 */
static inline u32 *
xdr_encode_nfstime4(u32 *p, nfstime4 time)
{
	p = xdr_encode_64(p,time.seconds);
	*p++ = 0;
	return p;
}

static inline u32 *
xdr_decode_fattr4_acl(u32 *p, fattr4_acl *facl)
{
/* dummy for now */
	return p;
}
static inline u32 *
xdr_decode_nfstime4_inplace(u32 *p)
{
	p = xdr_decode_64_inplace(p);
	*p = ntohl(*p);
	p++;
	return p;
}
static inline u32 *
xdr_decode_nfstime4(u32 *p, nfstime4 *timep)
{
	p = xdr_decode_64(p,&timep->seconds);
	p++;
	return p;
}

unsigned int bit4_mask[32] = {
        0x00000001,
        0x00000002,
        0x00000004,
        0x00000008,
        0x00000010,
        0x00000020,
        0x00000040,
        0x00000080,
        0x00000100,
        0x00000200,
        0x00000400,
        0x00000800,
        0x00001000,
        0x00002000,
        0x00004000,
        0x00008000,
        0x00010000,
        0x00020000,
        0x00040000,
        0x00080000,
        0x00100000,
        0x00200000,
        0x00400000,
        0x00800000,
        0x01000000,
        0x02000000,
        0x04000000,
        0x08000000,
        0x10000000,
        0x20000000,
        0x40000000,
        0x80000000
};


static u32 *
xdr_decode_bitmap4(u32 *p,bitmap4 *bm4)
{
int i;
uint32_t *ptr;

	bm4->BLEN = ntohl(*p++);
	nfsv4_printk(xdr,"NFSV4 xdr_decode_bitmap4: bits_len: %d\n",
				bm4->BLEN);
	for(i=0;i<bm4->BLEN;i++){
		if(i>=XDR_ATTR4_BLEN)
			return 0;
		ptr = &bm4->BVAL[i];
		*ptr=ntohl(*p++);
		nfsv4_printk(xdr,"NFSV4 xdr_decode_bitmap4: bits_val[%d]: 0x%x\n",
				i,bm4->BVAL[i]);
	}
	return p;
}

static u32*
xdr_encode_specdata4(u32 *p,fattr4_rawdev *rd)
{
	nfsv4_printk(xdr,"NFSV4 xdr_encode_specdata4  ");
	*p = htonl(rd->specdata1);
	nfsv4_printk(xdr,"specdata1 0x%x ",*p);
	p++;
	*p = htonl(rd->specdata2);
	nfsv4_printk(xdr,"specdata2 0x%x\n",*p);
	p++;
	return p;
}

static u32*
xdr_decode_specdata_inplace(u32 *p)
{
	*p = ntohl(*p++);
	*p = ntohl(*p++);
	return p;
}
static u32*
xdr_decode_specdata(u32 *p, fattr4_rawdev *rd)
{
	rd->specdata1 = ntohl(*p++);
	rd->specdata2 = ntohl(*p++);
	return p;
}
static u32*
xdr_encode_fsid4(u32 *p,fattr4_fsid *f4)
{
	p = xdr_encode_64(p,f4->major);
	p = xdr_encode_64(p,f4->minor);
	return p;
}
static u32*
xdr_decode_fsid4_inplace(u32 *p)
{
	p = xdr_decode_64_inplace(p);
	p = xdr_decode_64_inplace(p);
	return p;
}
static u32*
xdr_decode_fsid4(u32 *p, fattr4_fsid *f4)
{
	p = xdr_decode_64(p,&f4->major);
	p = xdr_decode_64(p,&f4->minor);
	return p;
}
static u32 *
xdr_encode_verifier4(u32 *p,verifier4 ch4)
{
	nfsv4_printk(verifier,"NFS4 encode_verifier. verifier:\n");
	print_verifier4(ch4);
	memcpy(p,ch4,NFS4_VERIFIER_SIZE);
	return p + XDR_QUADLEN(NFS4_VERIFIER_SIZE);
}
static inline u32 *
xdr_decode_verifier4(u32 *p, verifier4 v4p)
{
	nfsv4_printk(verifier,"NFSDV4: xdr_decode_verifier4\n");
	memcpy(v4p,(char *)p,NFS4_VERIFIER_SIZE);
	print_verifier4(v4p);
	return p + XDR_QUADLEN(NFS4_VERIFIER_SIZE);
}                                             

static u32*
xdr_encode_nfs_client_id4(u32 *p,nfs_client_id4 *id4)
{
	nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_client_id4 len %d  1\n",id4->id.id_len);
	p = xdr_encode_verifier4(p,id4->verifier);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_client_id4 2\n");
	*p++ = htonl(id4->id.id_len);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_client_id4 3\n");
	memcpy(p,id4->id.id_val,id4->id.id_len);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_client_id4 4\n");
	p += XDR_QUADLEN(id4->id.id_len);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_client_id4 5\n");
	return p;
}
static u32*
xdr_decode_nfs_client_id4(u32 *p,nfs_client_id4 *id4)
{
	p = xdr_decode_verifier4(p, id4->verifier);
	id4->id.id_len = ntohl(*p++);
	id4->id.id_val = (char *)p;
	p += XDR_QUADLEN(id4->id.id_len);
	return p;
}

static u32*
xdr_encode_clientaddr4(u32 *p,clientaddr4 *ca4)
{
int len = 0;	

	nfsv4_printk(xdr,"NFSV4 xdr_encode_clientaddr4 1\n");
	len = strlen(ca4->r_netid);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_clientaddr4 2\n");
	if(len > 1023)
		return NULL;
	*p++ = htonl(len);
	if(len != 0)
		memcpy(p,ca4->r_netid,len);
	nfsv4_printk(xdr,"NFSV4 xdr_encode_clientaddr4 3\n");
	p += XDR_QUADLEN(len); 

	len = strlen(ca4->r_addr);
	if(len > 1023)
		return NULL;
	*p++ = htonl(len);
	if(len != 0)
		memcpy(p,ca4->r_addr,len);
	p += XDR_QUADLEN(len); 

	nfsv4_printk(xdr,"NFSV4 xdr_encode_clientaddr4 4\n");
	return p;
}

static u32*
xdr_decode_clientaddr4(u32 *p,clientaddr4 *ca4)
{
int len = 0;
	len = ntohl(*p++);
	if(len > 1023)
		return NULL;
	ca4->r_netid = (char *)p;
	p += XDR_QUADLEN(len); 

	len = ntohl(*p++);
	if(len > 1023)
		return NULL;
	ca4->r_addr = (char *)p;
	p += XDR_QUADLEN(len); 

	return p;
}

static u32*
xdr_encode_cb_client4(u32 *p,cb_client4 *cb4)
{
	*p++ = htonl(cb4->cb_program); 
	nfsv4_printk(xdr,"NFSV4 xdr_encode_cb_client\n");
	p = xdr_encode_clientaddr4(p,&cb4->cb_location);
	return p;
}

static u32*
xdr_decode_cb_client4(u32 *p,cb_client4 *cb4)
{
	cb4->cb_program = ntohl(*p++); 
	p = xdr_decode_clientaddr4(p,&cb4->cb_location);
	return p;
}

static u32*
xdr_encode_fattr4(u32 *p,struct fattr4 *res)
{
bitmap4 *bm4 = &res->attrmask;
int i,j,bit;
void *cptr = res->FATTRSVAL;
void *end = (char *)cptr + res->FATTRSLEN;
u32 *start,*save;

	nfsv4_printk(attributes,"encode_fattr4 cptr 0x%x, end 0x%x, 
		res->FATTRSLEN %d\n", (unsigned int)cptr, 
		(unsigned int)end,res->FATTRSLEN);

	p = xdr_encode_bitmap4(p,bm4);
	save = p;
	
	/* encode the attrlist */
	*p++ = htonl(res->FATTRSLEN);
	start = p;
	nfsv4_printk(attributes,"NFSDV4: xdr_encode_fattr4: p, beginning. 0x%x\n",(unsigned int)p);
	nfsv4_printk(attributes,"NFSDV4: xdr_encode_fattr4: res->FATTRSLEN %d\n",
		res->FATTRSLEN);

	/* for each bit set in the bitmap */
	for(i=0;i<bm4->BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4->BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				nfsv4_printk(attributes,"NFSDV4: xdr_encode_fatttr4 bit %d is set\n",bit);
				switch(bit){
				case FATTR4_SUPPORTED_ATTRS:
				{		
				bitmap4 *b4p = (bitmap4 *)cptr;
				int size = sizeof(u_int) + (b4p->BLEN * sizeof(uint32_t));
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SUPPORTED_ATTRS\n");
					p = xdr_encode_bitmap4(p,(bitmap4 *)cptr);
					/* bitmap4_val = size 2 */
					if((cptr + size) > end) 
						goto out_overrun;
					cptr += size; 
				}
				break;
				case FATTR4_TYPE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TYPE\n");
					*p++ = htonl(*(fattr4_type *)cptr);
					if((cptr + sizeof(fattr4_type)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_type); 
				}
				break;
				case FATTR4_FH_EXPIRE_TYPE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FH_EXPIRE_TYPE\n");
					*p++ = htonl(*(fattr4_fh_expire_type *)cptr);
					if((cptr + sizeof(fattr4_fh_expire_type)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_fh_expire_type); 
				}
				break;
				case FATTR4_CHANGE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_CHANGE\n");
					p = xdr_encode_64(p,*(fattr4_change *)cptr);
					if((cptr + sizeof(fattr4_change)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_change);
				}
				break;
				case FATTR4_SIZE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SIZE\n");
					p = xdr_encode_64(p,*(fattr4_size *)cptr);
					if((cptr + sizeof(fattr4_size)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_size); 
				}
				break;
				case FATTR4_LINK_SUPPORT:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_LINK_SUPPORT\n");
					*p++ = htonl(*(fattr4_link_support *)cptr);
					if((cptr + sizeof(fattr4_link_support)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_link_support); 
				}
				break;
				case FATTR4_SYMLINK_SUPPORT:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SYMLINK_SUPPORT\n");
					*p++ = htonl(*(fattr4_symlink_support *)cptr);
					if((cptr + sizeof(fattr4_symlink_support)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_symlink_support); 
				}
				break;
				case FATTR4_NAMED_ATTR:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_NAMED_ATTR\n");
					*p++ = htonl(*(fattr4_named_attr *)cptr);
					if((cptr + sizeof(fattr4_named_attr)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_named_attr); 
				}
				break;
				case FATTR4_FSID:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FSID\n");
					p = xdr_encode_fsid4(p,(fattr4_fsid *)cptr);
					if((cptr + sizeof(fattr4_fsid)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_fsid); 
				}
				break;
				case FATTR4_UNIQUE_HANDLES:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_UNIQUE_HANDLES\n");
					*p++ = htonl(*(fattr4_unique_handles *)cptr);
					if((cptr + sizeof(fattr4_unique_handles)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_unique_handles); 
				}
				break;
				case FATTR4_LEASE_TIME:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_LEASE_TIME\n");
					*p++ = htonl(*(fattr4_lease_time *)cptr);
					if((cptr + sizeof(fattr4_lease_time)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_lease_time); 
				}
				break;
				case FATTR4_RDATTR_ERROR:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_RDATTR_ERROR\n");
					*p++ = htonl(*(fattr4_rdattr_error *)cptr);
					if((cptr + sizeof(fattr4_rdattr_error)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_rdattr_error); 
				}
				break;
				case FATTR4_ACL:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_ACL: NOT CODED!!!!!!!!!!!!!!!!\n");
				}
				break;
				case FATTR4_ACLSUPPORT:
				{
					nfsv4_printk(attributes,
						"NFSV4 xdr_encode_fattr4 FATTR4_ACLSUPPORT\n");
					*p++ = htonl(*(fattr4_aclsupport *)cptr);
					if((cptr + sizeof(fattr4_aclsupport)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_aclsupport); 
				}
				break;
				case FATTR4_ARCHIVE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_ARCHIVE\n");
					*p++ = htonl(*(fattr4_archive *)cptr);
					if((cptr + sizeof(fattr4_archive)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_archive); 
				}
				break;
				case FATTR4_CANSETTIME:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_CANSETTIME\n");
					*p++ = htonl(*(fattr4_cansettime *)cptr);
					if((cptr + sizeof(fattr4_cansettime)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_cansettime); 
				}
				break;
				case FATTR4_CASE_INSENSITIVE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_CASE_INSENSITIVE\n");
					*p++ = htonl(*(fattr4_case_insensitive *)cptr);
					if((cptr + sizeof(fattr4_case_insensitive)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_case_insensitive); 
				}
				break;
				case FATTR4_CASE_PRESERVING:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_CASE_PRESERVING\n");
					*p++ = htonl(*(fattr4_case_preserving *)cptr);
					if((cptr + sizeof(fattr4_case_preserving)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_case_preserving); 
				}
				break;
				case FATTR4_CHOWN_RESTRICTED:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_CHOWN_RESTRICTED\n");
					*p++ = htonl(*(fattr4_chown_restricted *)cptr);
					if((cptr + sizeof(fattr4_chown_restricted)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_chown_restricted);
				}
				break;
				case FATTR4_FILEHANDLE:
				{
				int len = (int)*(int *)cptr;
				int size = sizeof(u_int) + len;
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FILEHANDLE\n");
					if((cptr + size) > end) 
						goto out_overrun;
					if(!(p = xdr_encode_fh4(p,(nfs_fh4 *)cptr)))
						return NULL;
					else
						cptr += size; 
				}
				break;
				case FATTR4_FILEID:
				{
					nfsv4_printk(attributes,"NFSDV4 xdr_encode_fattr4: FATTR_FILEID = %ld\n",*(long int *)cptr);
					p = xdr_encode_64(p, *(fattr4_fileid *)cptr);
					if((cptr + sizeof(fattr4_fileid)) > end) 
						goto out_overrun;
					cptr = cptr + sizeof(fattr4_fileid);
				}
				break;
				case FATTR4_FILES_AVAIL:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FILES_AVAIL\n");
					p = xdr_encode_64(p,*(fattr4_files_avail *)cptr);
					if((cptr + sizeof(fattr4_files_avail)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_files_avail); 
				}
				break;
				case FATTR4_FILES_FREE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FILES_FREE\n");
					p = xdr_encode_64(p,*(fattr4_files_free *)cptr);
					if((cptr + sizeof(fattr4_files_free)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_files_free); 
				}
				break;
				case FATTR4_FILES_TOTAL:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_FILES_TOTAL\n");
					p = xdr_encode_64(p,*(fattr4_files_total *)cptr);
					if((cptr + sizeof(fattr4_files_total)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_files_total); 
				}
				break;
				case FATTR4_FS_LOCATIONS:
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 not CODED!  FATTR4_FS_LOCATIONS\n");
				break;
				case FATTR4_HIDDEN:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_HIDDEN\n");
					*p++ = htonl(*(fattr4_hidden *)cptr);
					if((cptr + sizeof(fattr4_hidden)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_hidden);; 
				}
				break;
				case FATTR4_HOMOGENEOUS:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_HOMOGENEOUS\n");
					*p++ = htonl(*(fattr4_homogeneous *)cptr);
					if((cptr + sizeof(fattr4_homogeneous)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_homogeneous); 
				}
				break;
				case FATTR4_MAXFILESIZE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MAXFILESIZE\n");
					p = xdr_encode_64(p,*(fattr4_maxfilesize *)cptr);
					if((cptr + sizeof(fattr4_maxfilesize)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_maxfilesize); 
				}
				break;
				case FATTR4_MAXLINK:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MAXLINK\n");
					*p++ = htonl(*(fattr4_maxlink *)cptr);
					if((cptr + sizeof(fattr4_maxlink)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_maxlink);
				}
				break;
				case FATTR4_MAXNAME:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MAXNAME\n");
					*p++ = htonl(*(fattr4_maxname *)cptr);
					if((cptr + sizeof(fattr4_maxname)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_maxname);
				}
				break;
				case FATTR4_MAXREAD:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MAXREAD\n");
					p = xdr_encode_64(p,*(fattr4_maxread *)cptr);
					if((cptr + sizeof(fattr4_maxread)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_maxread);
				}
				break;
				case FATTR4_MAXWRITE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MAXWRITE\n");
					p = xdr_encode_64(p,*(fattr4_maxwrite *)cptr);
					if((cptr + sizeof(fattr4_maxwrite)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_maxwrite);
				}
				break;
				case FATTR4_MIMETYPE:
					printk("NFSV4 xdr_encode_fattr4 FATTR4_MAXWRITE NOT CODED !!!!!!!!!!!!!!!!\n");
				break;
				case FATTR4_MODE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MODE\n");
					*p++ = htonl(*(fattr4_mode *)cptr);
					if((cptr + sizeof(fattr4_mode)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_mode);
				}
				break;
				case FATTR4_NO_TRUNC:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_NO_TRUNC\n");
					*p++ = htonl(*(fattr4_no_trunc *)cptr);
					if((cptr + sizeof(fattr4_no_trunc)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_no_trunc);
				}
				break;
				case FATTR4_NUMLINKS:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_NUMLINKS\n");
					*p++ = htonl(*(fattr4_numlinks *)cptr);
					if((cptr + sizeof(fattr4_numlinks)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_numlinks);
				}
				break;
				case FATTR4_OWNER:
				{
				u_int len;
				fattr4_owner *ptr = (fattr4_owner *)cptr;

					len = ptr->utf8string_len;
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_OWNER: len %d \n",len);
					p = xdr_encode_utf8string(p,ptr);
					if((cptr + sizeof(fattr4_owner) + len ) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_owner) +len;
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
				u_int len;
				fattr4_owner_group *ptr = (fattr4_owner_group*)cptr;

					len = ptr->utf8string_len;
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_OWNER_GROUP: len %d\n",len);
					p = xdr_encode_utf8string(p,ptr);
					if((cptr + sizeof(fattr4_owner_group) + len ) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_owner_group) +len;
				}
				break;
				case FATTR4_QUOTA_HARD:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_HARD\n");
					p = xdr_encode_64(p,*(fattr4_quota_hard *)cptr);
					if((cptr + sizeof(fattr4_quota_hard)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_quota_hard);
				}
				break;
				case FATTR4_QUOTA_SOFT:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_SOFT\n");
					p = xdr_encode_64(p,*(fattr4_quota_soft *)cptr);
					if((cptr + sizeof(fattr4_quota_soft)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_quota_soft);
				}
				break;
				case FATTR4_QUOTA_USED:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_USED\n");
					p = xdr_encode_64(p,*(fattr4_quota_used *)cptr);
					if((cptr + sizeof(fattr4_quota_used)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_quota_used);
				}
				break;
				case FATTR4_RAWDEV:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_RAWDEV\n");
					p = xdr_encode_specdata4(p,(fattr4_rawdev *)cptr);
					if((cptr + sizeof(fattr4_rawdev)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_rawdev);
				}
				break;
				case FATTR4_SPACE_AVAIL:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SPACE_AVAIL\n");
					p = xdr_encode_64(p,*(fattr4_space_avail *)cptr);
					if((cptr + sizeof(fattr4_space_avail)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_space_avail);
				}
				break;
				case FATTR4_SPACE_FREE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SPACE_FREE\n");
					p = xdr_encode_64(p,*(fattr4_space_free *)cptr);
					if((cptr + sizeof(fattr4_space_free)) > end) 
						goto out_overrun;
					cptr +=sizeof(fattr4_space_free);
				}
				break;
				case FATTR4_SPACE_TOTAL:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SPACE_TOTAL\n");
					p = xdr_encode_64(p,*(fattr4_space_total *)cptr);
					if((cptr + sizeof(fattr4_space_total)) > end) 
						goto out_overrun;
					cptr +=sizeof(fattr4_space_total);
				}
				break;
				case FATTR4_SPACE_USED:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SPACE_USED\n");
					p = xdr_encode_64(p,*(fattr4_space_used *)cptr);
					if((cptr + sizeof(fattr4_space_used)) > end) 
						goto out_overrun;
					cptr +=sizeof(fattr4_space_used);
				}
				break;
				case FATTR4_SYSTEM:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_SYSTEM\n");
					*p++ = htonl(*(fattr4_system *)cptr);
					if((cptr + sizeof(fattr4_system)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_system);
				}
				break;
				case FATTR4_TIME_ACCESS:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_ACCESS\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_access *)cptr);
					if((cptr + sizeof(fattr4_time_access)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_access);
				}
				break;
				case FATTR4_TIME_ACCESS_SET:
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_ACCESS_SET\n");
				break;
				case FATTR4_TIME_BACKUP:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_BACKUP\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_backup *)cptr);
					if((cptr + sizeof(fattr4_time_backup)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_backup);
				}
				break;
				case FATTR4_TIME_CREATE:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_CREATE\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_create *)cptr);
					if((cptr + sizeof(fattr4_time_create)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_create);
				}
				break;
				case FATTR4_TIME_DELTA:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_DELTA\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_delta *)cptr);
					if((cptr + sizeof(fattr4_time_delta)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_delta);
				}
				break;
				case FATTR4_TIME_METADATA:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_METADATA\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_metadata *)cptr);
					if((cptr + sizeof(fattr4_time_metadata)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_metadata);
				}
				break;
				case FATTR4_TIME_MODIFY:
				{
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_MODIFY:\n");
					p = xdr_encode_nfstime4(p,*(fattr4_time_modify *)cptr);
					nfsv4_printk(attributes,"encode_fattr4 cptr 0x%x time_modify %d\n",
						(unsigned int)cptr,sizeof(fattr4_time_modify));
					if((cptr + sizeof(fattr4_time_modify)) > end) 
						goto out_overrun;
					cptr += sizeof(fattr4_time_modify);
				}
				break;
				case FATTR4_TIME_MODIFY_SET:
					nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_TIME_MODIFY_SET:\n");
				break;
				default:
				break;
				}
			}
		}
	}
	nfsv4_printk(attributes,"NFSDV4: xdr_encode_fattr4: p, end. 0x%x\n and with XDR_QUADLEN(res->FATTRSLEN): 0x%x\n",(unsigned int)p,(unsigned int)p + XDR_QUADLEN(res->FATTRSLEN));
	*save = htonl((p-start)*4);
	return p;
out_overrun:
	nfsv4_printk(attributes,"NFSDV4: xdr_encode_fattr4: OVERRUN, return NULL\n");
	return NULL;
}
static u32 *
xdr_decode_bitmap4_inplace(u32 *p,bitmap4 *b4)
{
int i;
	*p = ntohl(*p);
	b4->BLEN = *p;
	p++;
	b4->BVAL = p;
	for(i=0;i<b4->BLEN;i++){
		*p =ntohl(*p);
		p++;
	}
	return p;
}

u32 *
xdr_decode_fattr4_inplace(u32 *p)
{
int i,j;
int bit,fattrslen;
bitmap4 bm4 ;

	nfsv4_printk(readdir,"NFSV4 xdr_decode_fattr_inplace: calling decode bitmap4\n");
	p = xdr_decode_bitmap4_inplace(p,&bm4);
	nfsv4_printk(readdir,"           bitmap_len: %d\n",bm4.BLEN); 
	nfsv4_printk(readdir,"           bitmap_val[0]: 0x%x\n",bm4.BVAL[0]);
	nfsv4_printk(readdir,"           bitmap_val[1]: 0x%x\n",bm4.BVAL[1]);

	fattrslen = *p = ntohl(*p);
	p++;
	nfsv4_printk(readdir,"NFSV4 xdr_decode_fattr4_inplace: attrs_len = %d p = 0x%x\n",fattrslen,(unsigned int)p);
	/* for each bit set in the bit mask */
	for(i=0;i<bm4.BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4.BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				nfsv4_printk(readdir,"NFSV4 xdr_decode_fattr_inplace:  bit %d set\n", bit);
				switch(bit){
				case FATTR4_SUPPORTED_ATTRS:
				{
				bitmap4 bb;
					p = xdr_decode_bitmap4_inplace(p,&bb);
				}
				break;
				case FATTR4_TYPE:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_FH_EXPIRE_TYPE:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_CHANGE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_SIZE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_LINK_SUPPORT:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_SYMLINK_SUPPORT:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_NAMED_ATTR:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_FSID:
				{
					p = xdr_decode_fsid4_inplace(p);
				}
				break;
				case FATTR4_UNIQUE_HANDLES:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_LEASE_TIME:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_RDATTR_ERROR:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_ACL:
				{
					/* NOTE: NOT DONE YET
					p = xdr_decode_fattr4_acl(p,(fattr4_acl *)p);
					*/
				}
				break;
				case FATTR4_ACLSUPPORT:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_ARCHIVE:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_CANSETTIME:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_CASE_INSENSITIVE:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_CASE_PRESERVING:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_CHOWN_RESTRICTED:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_FILEHANDLE:
				{
					p = xdr_decode_nfs_fh4_inplace(p);
				}
				break;
				case FATTR4_FILEID:
				{
				u64 *ptr = (u64 *)p;
					p = xdr_decode_64_inplace(p);
					nfsv4_printk(readdir,"NFSV4 xdr_decode_fattr4_inplace : FILEID: %ld\n", *(long int*)ptr);
				}
				break;
				case FATTR4_FILES_AVAIL:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_FILES_FREE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_FILES_TOTAL:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_FS_LOCATIONS:
				/* NOT Implemtented */
				break;
				case FATTR4_HIDDEN:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_HOMOGENEOUS:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_MAXFILESIZE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_MAXLINK:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_MAXNAME:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_MAXREAD:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_MAXWRITE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_MIMETYPE:
				/* NOT Implemented */
				break;
				case FATTR4_MODE:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_NO_TRUNC:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_NUMLINKS:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_OWNER:
				{
				  utf8string str;
					str.utf8string_len = *p = ntohl(*p);
					p++;
					str.utf8string_val = (char *)p;
					p += XDR_QUADLEN(str.utf8string_len);
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
				utf8string str;
					str.utf8string_len = *p = ntohl(*p);
					p++;
					str.utf8string_val = (char *)p;
					p += XDR_QUADLEN(str.utf8string_len);
				}
				break;
				case FATTR4_QUOTA_HARD:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_QUOTA_SOFT:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_QUOTA_USED:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_RAWDEV:
				{
					p = xdr_decode_specdata_inplace(p);
				}
				break;
				case FATTR4_SPACE_AVAIL:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_SPACE_FREE:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_SPACE_TOTAL:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_SPACE_USED:
				{
					p = xdr_decode_64_inplace(p);
				}
				break;
				case FATTR4_SYSTEM:
				{
					*p = ntohl(*p);
					p++;
				}
				break;
				case FATTR4_TIME_ACCESS:
				{
					p = xdr_decode_nfstime4_inplace(p);
				}
				case FATTR4_TIME_ACCESS_SET:
					printk("NFSV4 FATTR4_TIME_ACCESS_SET: not coded\n");
				break;
				case FATTR4_TIME_BACKUP:
				{
					p = xdr_decode_nfstime4_inplace(p);
				}
				break;
				case FATTR4_TIME_CREATE:
				{
					p = xdr_decode_nfstime4_inplace(p);
				}
				break;
				case FATTR4_TIME_DELTA:
				{
					p = xdr_decode_nfstime4_inplace(p);
				}
				break;
				case FATTR4_TIME_METADATA:
				{
					p = xdr_decode_nfstime4_inplace(p);
				}
				break;
				case FATTR4_TIME_MODIFY:
				{
					nfsv4_printk(readdir, "NFSV4 xdr_decode_fattr4_inplace TIME MODIFY\n");
					p = xdr_decode_nfstime4_inplace(p);
				}
				case FATTR4_TIME_MODIFY_SET:
					nfsv4_printk(readdir, "NFSV4 xdr_decode_fattr4_inplace TIME_MODIFY_SET\n");
				break;
				default:
				break;
				}
			}
		}
	}
	nfsv4_printk(readdir,"NFSV4: xdr_decode_fattr4_inplace: end. p = 0x%x\n",(unsigned int)p);
	return p;
}

/*
 *  preallocate res->FATTRSVAL before calling
 */

u32 *
xdr_decode_fattr4(u32 *p,fattr4 *res)
{
int i,j;
int bit;
bitmap4 *bm4 = &res->attrmask;
void *cptr = res->FATTRSVAL;

	nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr: calling decode bitmap4\n");
	p = xdr_decode_bitmap4(p,bm4);
	res->FATTRSLEN = ntohl(*p++);
	nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4: attrs_len = %d p = 0x%x\n",res->FATTRSLEN,(unsigned int)p);
	/* for each bit set in the bit mask */
	for(i=0;i<bm4->BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4->BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr:  bit %d set\n", bit);
				switch(bit){
				case FATTR4_SUPPORTED_ATTRS:
				{
					p = xdr_decode_bitmap4(p,(bitmap4 *)cptr);
				/* NOTE: get rid of hardwired 2 */
					cptr += sizeof(bitmap4) + 2;
				}
				break;
				case FATTR4_TYPE:
				{
					*(fattr4_type *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_type);
				}
				break;
				case FATTR4_FH_EXPIRE_TYPE:
				{
					*(fattr4_fh_expire_type *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_fh_expire_type);
				}
				break;
				case FATTR4_CHANGE:
				{
					p = xdr_decode_64(p,(fattr4_change *)cptr);
					cptr += sizeof(fattr4_change);
				}
				break;
				case FATTR4_SIZE:
				{
					p = xdr_decode_64(p,(fattr4_size *)cptr);
					cptr += sizeof(fattr4_size);
				}
				break;
				case FATTR4_LINK_SUPPORT:
				{
					*(fattr4_link_support *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_link_support);
				}
				break;
				case FATTR4_SYMLINK_SUPPORT:
				{
					*(fattr4_symlink_support *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_symlink_support);
				}
				break;
				case FATTR4_NAMED_ATTR:
				{
					*(fattr4_named_attr *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_named_attr);
				}
				break;
				case FATTR4_FSID:
				{
					p = xdr_decode_fsid4(p,(fattr4_fsid *)cptr);
					cptr += sizeof(fattr4_fsid);
				}
				break;
				case FATTR4_ACL:
				{
					/* NOTE: NOT DONE YET
					p = xdr_decode_fattr4_acl(p,(fattr4_acl *)cptr);
					cptr += sizeof(fattr4_acl);
					*/
				}
				break;
				case FATTR4_ARCHIVE:
				{
					*(fattr4_archive *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_archive);
				}
				break;
				case FATTR4_CANSETTIME:
				{
					*(fattr4_cansettime *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_cansettime);
				}
				break;
				case FATTR4_CASE_INSENSITIVE:
				{
					*(fattr4_case_insensitive *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_case_insensitive);
				}
				break;
				case FATTR4_CASE_PRESERVING:
				{
					*(fattr4_case_preserving *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_case_preserving);
				}
				break;
				case FATTR4_CHOWN_RESTRICTED:
				{
					*(fattr4_chown_restricted *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_chown_restricted);
				}
				break;
				case FATTR4_FILEHANDLE:
				{
					p = xdr_decode_nfs_fh4(p,(nfs_fh4 *)cptr);
					cptr += sizeof(nfs_fh4);
				}
				break;
				case FATTR4_FILEID:
				{
					p = xdr_decode_64(p,(fattr4_fileid *)cptr);
					nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4 : FILEID: %ld\n", *(long int*)cptr);
					cptr += sizeof(fattr4_fileid);
				}
				break;
				case FATTR4_FILES_AVAIL:
				{
					p = xdr_decode_64(p,(fattr4_files_avail *)cptr);
					cptr += sizeof(fattr4_files_avail);
				}
				break;
				case FATTR4_FILES_FREE:
				{
					p = xdr_decode_64(p,(fattr4_files_free *)cptr);
					cptr += sizeof(fattr4_files_free);
				}
				break;
				case FATTR4_FILES_TOTAL:
				{
					p = xdr_decode_64(p,(fattr4_files_total *)cptr);
					cptr += sizeof(fattr4_files_total);
				}
				break;
				case FATTR4_HIDDEN:
				{
					*(fattr4_hidden *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_hidden);
				}
				break;
				case FATTR4_HOMOGENEOUS:
				{
					*(fattr4_homogeneous *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_homogeneous);
				}
				break;
				case FATTR4_MAXFILESIZE:
				{
					p = xdr_decode_64(p,(fattr4_maxfilesize *)cptr);
					cptr += sizeof(fattr4_maxfilesize);
				}
				break;
				case FATTR4_MAXLINK:
				{
					*(fattr4_maxlink *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_maxlink);
				}
				break;
				case FATTR4_MAXNAME:
				{
					*(fattr4_maxname *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_maxname);
				}
				break;
				case FATTR4_MAXREAD:
				{
					p = xdr_decode_64(p,(fattr4_maxread *)cptr);
					cptr += sizeof(fattr4_maxread);
				}
				break;
				case FATTR4_MAXWRITE:
				{
					p = xdr_decode_64(p,(fattr4_maxwrite *)cptr);
					cptr += sizeof(fattr4_maxwrite);
				}
				break;
				case FATTR4_MODE:
				{
					*(fattr4_mode *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_mode);
				}
				break;
				case FATTR4_NO_TRUNC:
				{
					*(fattr4_no_trunc *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_no_trunc);
				}
				break;
				case FATTR4_NUMLINKS:
				{
					*(fattr4_numlinks *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_numlinks); 
				}
				break;
				case FATTR4_OWNER:
				{
				  int l;

				  *((u_int *) cptr) = l = ntohl(*p++);
				  cptr += sizeof(u_int);
				  *((char **) cptr) = cptr + sizeof(char *);
				  cptr += sizeof(char *);
				  memcpy (cptr, p, l);
				  cptr += l;
				  p += XDR_QUADLEN(l);
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
				  int l;

				  *((u_int *) cptr) = l = ntohl(*p++);
				  cptr += sizeof(u_int);
				  *((char **) cptr) = cptr + sizeof(char *);
				  cptr += sizeof(char *);
				  memcpy (cptr, p, l);
				  cptr += l;
				  p += XDR_QUADLEN(l);
				}
				break;
				case FATTR4_QUOTA_HARD:
				{
					p = xdr_decode_64(p,(fattr4_quota_hard *)cptr);
					cptr += sizeof(fattr4_quota_hard);
				}
				break;
				case FATTR4_QUOTA_SOFT:
				{
					p = xdr_decode_64(p,(fattr4_quota_soft *)cptr);
					cptr += sizeof(fattr4_quota_soft);
				}
				break;
				case FATTR4_QUOTA_USED:
				{
					p = xdr_decode_64(p,(fattr4_quota_used *)cptr);
					cptr += sizeof(fattr4_quota_used);
				}
				break;
				case FATTR4_RAWDEV:
				{
					p = xdr_decode_specdata(p,(fattr4_rawdev *)cptr);
					cptr += sizeof(fattr4_rawdev);
				}
				break;
				case FATTR4_SPACE_AVAIL:
				{
					p = xdr_decode_64(p,(fattr4_space_avail *)cptr);
					cptr += sizeof(fattr4_space_avail);
				}
				break;
				case FATTR4_SPACE_FREE:
				{
					p = xdr_decode_64(p,(fattr4_space_free *)cptr);
					cptr += sizeof(fattr4_space_free);
				}
				break;
				case FATTR4_SPACE_TOTAL:
				{
					p = xdr_decode_64(p,(fattr4_space_total *)cptr);
					cptr += sizeof(fattr4_space_total);
				}
				break;
				case FATTR4_SPACE_USED:
				{
					p = xdr_decode_64(p,(fattr4_space_used *)cptr);
					cptr += sizeof(fattr4_space_used);
				}
				break;
				case FATTR4_SYSTEM:
				{
					*(fattr4_system *)cptr = ntohl(*p++);
					cptr += sizeof(fattr4_system);
				}
				break;
				case FATTR4_TIME_ACCESS:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_access *)cptr);
					cptr += sizeof(fattr4_time_access);
				}
				break;
				case FATTR4_TIME_BACKUP:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_backup *)cptr);
					cptr += sizeof(fattr4_time_backup);
				}
				break;
				case FATTR4_TIME_CREATE:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_create *)cptr);
					cptr += sizeof(fattr4_time_create);
				}
				break;
				case FATTR4_TIME_DELTA:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_delta *)cptr);
					cptr += sizeof(fattr4_time_delta);
				}
				break;
				case FATTR4_TIME_METADATA:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_metadata *)cptr);
					cptr += sizeof(fattr4_time_metadata);
				}
				break;
				case FATTR4_TIME_MODIFY:
				{
					p = xdr_decode_nfstime4(p,(fattr4_time_modify *)cptr);
					cptr += sizeof(fattr4_time_modify);
				}
				break;
				default:
				break;
				}
			}
		}
	}
	nfsv4_printk(attributes,"NFSV4: xdr_decode_fattr4: end. p = 0x%x\n",(unsigned int)p);
#if 0
	print_fattr4(res);
#endif /*0*/
	return p;
}

static u32 *
xdr_decode_write4(u32 *p,WRITE4res *res)
{
	nfsv4_printk(write,"NFSV4 xdr_decode_write4\n");
	res->WRITERES.count = ntohl(*p++);
	res->WRITERES.committed = ntohl(*p++);
	p = xdr_decode_verifier4(p,res->WRITERES.writeverf);
	return p;
}

static u32 *
xdr_decode_read4(u32 *p,READ4res *res)
{
	res->READLEN = ntohl(*p++);	

	nfsv4_printk(read,"NFSV4 xdr_decode_read4 READLEN: %d \n",res->READLEN);
	return p + XDR_QUADLEN(res->READLEN);
}

/* decode in place */
static u32 *
xdr_decode_entry4(u32 *p)
{
utf8string str;
	nfsv4_printk(readdir,"NFSV4 xdr_decode_entry4!!!!\n");

	/* nextentry */
	*p = ntohl(*p);
	nfsv4_printk(readdir,"NFSV4 xdr_decode_entry4 nextentry %d\n",*p);
	p++;

	/* cookie */
	p = xdr_decode_64_inplace(p);
	
	/* utf8string*/
	*p = ntohl(*p);	
	nfsv4_printk(readdir,"name_len %d\n",*p);
	str.utf8string_len = *p++; 
	str.utf8string_val = (char *)p;
	nfsv4_printk(xdr,"        name: ");
	nfsv4_printk(readdir,"name_char (first) %c\n",(char) *p);
	print_utf8string(&str);
	p +=  XDR_QUADLEN(str.utf8string_len); 
	p = xdr_decode_fattr4_inplace(p);
	return p;
}

static u32 *
xdr_decode_dirlist4(struct rpc_rqst *req,u32 *p,dirlist4 *res)
{
	int next;
	u32 *p1;

	/* on the wire, entry4.nextentry is first. */
	p1 = req->rq_rvec[1].iov_base;
	next = ntohl(*p1);	
	nfsv4_printk(readdir,"NFSV4 xdr_decode_dirlist4. next %d\n",next);

	while(next){
		p1 = xdr_decode_entry4(p1);
		next = ntohl(*p1);
		nfsv4_printk(readdir,"NFSV4 xdr_decode_dirlist next %d \n",next);	
	}
	p1++;
	res->eof = ntohl(*p1++);
	nfsv4_printk(readdir,"NFSV4 xdr_decode_dirlist res->eof %d\n",res->eof);	
	return p;
}

/*
 * NFS encode functions
 */
static u32 *
xdr_encode_write4data(u32 *p,WRITE4args *args)
{
	char prntbuf[32];

	nfsv4_printk(write,"NFSV4 xdr_encode_write4data: WRITELEN %d\n",args->WRITELEN);
	*p++ = htonl(args->WRITELEN);	

#ifndef __OpenBSD__
	memset(prntbuf,0,32);
	memcpy(prntbuf,(u8 *)args->WRITEVAL,
		(args->WRITELEN <= 31?args->WRITELEN:31));
	nfsv4_printk(write,"NFSV4 xdr_encode_write4data WRITELEN: %d first 31 bytes: %s\n",args->WRITELEN,prntbuf);

	/* kmalloc by caller - this kmalloc and copy will 
         * go away once iovec is used.
	 */	
	memcpy((u8 *)p,args->WRITEVAL,args->WRITELEN);
	p += XDR_QUADLEN(args->WRITELEN);
#endif
	
	return p;
}
static u32*
xdr_encode_pathname4(u32 *p,pathname4 *p4)
{
int j;
	*p++ = htonl(p4->pathname4_len);
	for(j=0;j<p4->pathname4_len;j++) 
		p = xdr_encode_utf8string(p,&p4->pathname4_val[j]);
	return p;
}


static u32*
xdr_encode_createtype4(u32 *p, createtype4 *objt)
{
	*p++ = htonl(objt->type);
	switch(objt->type){
	case NF4LNK:
		/* link data */
		p = xdr_encode_utf8string(p, &objt->createtype4_u.linkdata);
	break;
	case NF4BLK:
	case NF4CHR:
		/* device data */
		*p++ = htonl(objt->createtype4_u.devdata.specdata1);
		*p++ = htonl(objt->createtype4_u.devdata.specdata2);
	break;
	case NF4SOCK:
	case NF4FIFO:
	case NF4DIR:
		/* void */
	break;
	default:
		nfsv4_printk(xdr, "NFSV4: xdr_encode_createtype4: invalid object type %d\n", objt->type);
	break;
	}
	return p;
}
static u32 *
xdr_encode_createhow4(u32 *p,createhow4 *ch4)
{
	*p++ = htonl(ch4->mode);
	switch(ch4->mode){
	case UNCHECKED4:
	case GUARDED4:
		p = xdr_encode_fattr4(p,&ch4->createhow4_u.createattrs);
	break;
case EXCLUSIVE4:
		p = xdr_encode_verifier4(p,ch4->createhow4_u.createverf);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_encode_createhow4: invlaid mode %d\n",ch4->mode);
	}
	return p;
}
static u32 *
xdr_encode_nfs_lockowner4(u32 *p,nfs_lockowner4 *lown)
{
	p = xdr_encode_64(p,lown->clientid);
	*p++ = htonl(lown->owner.owner_len);
	memcpy(p,lown->owner.owner_val,lown->owner.owner_len);
	print_lockowner(lown);
	return p + XDR_QUADLEN(lown->owner.owner_len);
}
static u32 *
xdr_decode_nfs_lockowner4(u32 *p,nfs_lockowner4 *lown)
{
	p = xdr_decode_64(p,&lown->clientid);
	lown->owner.owner_len = ntohl(*p++);
	lown->owner.owner_val = (char *)p;
	return p + XDR_QUADLEN(lown->owner.owner_len);
}


static u32 *
xdr_encode_openflag4(u32 *p,openflag4 *of4)
{
	*p++ = htonl(of4->opentype);
	switch (of4->opentype){
	case OPEN4_CREATE: 
		p = xdr_encode_createhow4(p,&of4->openflag4_u.how);
	break;
	case OPEN4_NOCREATE:
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_encode_openflag4: invlaid switch %d\n",of4->opentype);
	break;
	}
	return p;
}
static u32 *
xdr_encode_open_claim_delegate_cur4(u32 *p,open_claim_delegate_cur4  *cur4)
{
	p = xdr_encode_pathname4(p,&cur4->file);
	p = xdr_encode_64(p,cur4->delegate_stateid);
	return p;
}
static u32 *
xdr_encode_open_claim4(u32 *p,open_claim4  *c4)
{
	*p++ = htonl(c4->claim);	
	switch(c4->claim){
	case CLAIM_NULL:
		p = xdr_encode_pathname4(p,&c4->open_claim4_u.file);
	break;
	case CLAIM_PREVIOUS:
		*p++ = htonl(c4->open_claim4_u.delegate_type);
	break;
	case CLAIM_DELEGATE_CUR:
		p = xdr_encode_open_claim_delegate_cur4(p,&c4->open_claim4_u.delegate_cur_info);
	break;
	case CLAIM_DELEGATE_PREV:
		p = xdr_encode_pathname4(p,&c4->open_claim4_u.file_delegate_prev);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_encode_open_claim4: invlaid claim %d\n",c4->claim);
	break;
	}
	return p;
}

static u32*
xdr_encode_change_info4(u32 *p, change_info4 *c4)
{
	*p++= htonl(c4->atomic);
	p = xdr_encode_64(p, c4->before);
	p = xdr_encode_64(p, c4->after);
	return p;
}

static u32*
xdr_decode_change_info4(u32 *p, change_info4 *c4)
{
	c4->atomic = ntohl(*p++);
	p = xdr_decode_64(p, &c4->before);
	p = xdr_decode_64(p, &c4->after);
	return p;
}

static u32*
xdr_decode_nfs_modified_limit4(u32 *p,nfs_modified_limit4 *m4)
{
	m4->num_blocks = ntohl(*p++);
	m4->bytes_per_block = ntohl(*p++);
	return p;
}

static u32*
xdr_encode_nfs_modified_limit4(u32 *p,nfs_modified_limit4 *m4)
{
	*p++ = htonl(m4->num_blocks);
	*p++ = htonl(m4->bytes_per_block);
	return p;
}
static u32*
xdr_decode_nfs_space_limit4(u32 *p, nfs_space_limit4 *s4)
{
	s4->limitby = ntohl(*p++);
	switch(s4->limitby){
	case NFS_LIMIT_SIZE:
		p = xdr_decode_64(p,&s4->nfs_space_limit4_u.filesize);
	break;
	case NFS_LIMIT_BLOCKS:
		p = xdr_decode_nfs_modified_limit4(p,&s4->nfs_space_limit4_u.mod_blocks);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_nfs_space_limit4: invalid switch %d\n",s4->limitby);
	break;
	}
	return p;
}

static u32*
xdr_encode_nfs_space_limit4(u32 *p, nfs_space_limit4 *s4)
{
	*p++ = htonl(s4->limitby);
	switch(s4->limitby){
	case NFS_LIMIT_SIZE:
		p = xdr_encode_64(p,s4->nfs_space_limit4_u.filesize);
	break;
	case NFS_LIMIT_BLOCKS:
		p = xdr_encode_nfs_modified_limit4(p,&s4->nfs_space_limit4_u.mod_blocks);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_space_limit4: invalid switch %d\n",s4->limitby);
	break;
	}
	return p;
}

static u32*
xdr_decode_nfsace4(u32 *p, nfsace4* acl)
{
	acl->type = ntohl(*p++);
	acl->flag = ntohl(*p++);
	acl->access_mask = ntohl(*p++);
	p = xdr_decode_utf8string(p,&acl->who);
	return p;
}

static u32*
xdr_encode_nfsace4(u32 *p, nfsace4* acl)
{
	*p++ = htonl(acl->type);
	*p++ = htonl(acl->flag);
	*p++ = htonl(acl->access_mask);
	p = xdr_encode_utf8string(p, &acl->who);
	return p;
}

static u32*
xdr_decode_open_write_delegation4(u32 *p, open_write_delegation4 *w4)
{
	p = xdr_decode_64(p, &w4->stateid);
	w4->recall = ntohl(*p++);
	p  = xdr_decode_nfs_space_limit4(p, &w4->space_limit);
	p = xdr_decode_nfsace4(p, &w4->permissions);
	return p;
}

static u32*
xdr_encode_open_write_delegation4(u32 *p, open_write_delegation4 *w4)
{
	p = xdr_encode_64(p, w4->stateid);
	*p++ = htonl(w4->recall);
	p  = xdr_encode_nfs_space_limit4(p, &w4->space_limit);
	p = xdr_encode_nfsace4(p, &w4->permissions);
	return p;
}
static u32*
xdr_decode_open_read_delegation4(u32 *p, open_read_delegation4 *r4)
{
	p = xdr_decode_64(p, &r4->stateid);
	r4->recall = ntohl(*p++);
	p = xdr_decode_nfsace4(p, &r4->permissions);
	return p;
}

static u32*
xdr_encode_open_read_delegation4(u32 *p, open_read_delegation4 *r4)
{
	p = xdr_encode_64(p, r4->stateid);
	*p++ = htonl(r4->recall);
	p = xdr_encode_nfsace4(p, &r4->permissions);
	return p;
}
static u32*
xdr_decode_open_delegation4(u32 *p, open_delegation4 *d4)
{
	d4->delegation_type = ntohl(*p++);
	switch(d4->delegation_type){
	case OPEN_DELEGATE_NONE:
	break;
	case OPEN_DELEGATE_READ:
		p = xdr_decode_open_read_delegation4(p,&d4->open_delegation4_u.read);
	break;
	case OPEN_DELEGATE_WRITE:
	p = xdr_decode_open_write_delegation4(p,&d4->open_delegation4_u.write);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_open_delegation4: invalid type %d\n",d4->delegation_type);
	break;
	}
	return p;
}
static u32*
xdr_encode_open_delegation4(u32 *p, open_delegation4 *d4)
{
	*p++ = htonl(d4->delegation_type);
	switch(d4->delegation_type){
	case OPEN_DELEGATE_NONE:
	break;
	case OPEN_DELEGATE_READ:
		p = xdr_encode_open_read_delegation4(p,&d4->open_delegation4_u.read);
	break;
	case OPEN_DELEGATE_WRITE:
		p = xdr_encode_open_write_delegation4(p,&d4->open_delegation4_u.write);
	break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_open_delegation4: invalid type %d\n",d4->delegation_type);
	break;
	}
	return p;
}

static u32*
xdr_decode_sec_oid4(u32 *p, sec_oid4 *s4)
{
	s4->sec_oid4_len = ntohl(*p++);
	nfsv4_printk(secinfo,"NFSV4 xdr_decode_sec_oid4 sec_oid4_len %d\n",s4->sec_oid4_len);
	s4->sec_oid4_val = (char *)p;
	return p + XDR_QUADLEN(s4->sec_oid4_len);
}

static u32*
xdr_decode_secinfo4(u32 *p, secinfo4 *info)
{
int j;
	info->flavor = ntohl(*p++);
	nfsv4_printk(secinfo,"NFSV4: xdr_decode_secinfo4 flavor %d\n", info->flavor);
	switch(info->flavor){
	case RPC_AUTH_NULL:
	case RPC_AUTH_UNIX:
	break;
	case RPC_AUTH_GSS: /*6*/
	{
	rpcsec_gss_info *rpcinfo;
	int size;
		info->flavor_info.flavor_info_len = ntohl(*p++);
		nfsv4_printk(secinfo,"NFSV4: xdr_decode_secinfo4 flavor_info_len %d\n", 
			info->flavor_info.flavor_info_len);
		size = info->flavor_info.flavor_info_len * sizeof(rpcsec_gss_info);
		info->flavor_info.flavor_info_val = kmalloc(size,GFP_KERNEL);
		memset(info->flavor_info.flavor_info_val,0,size);

		rpcinfo = (rpcsec_gss_info *)&info->flavor_info.flavor_info_val[0];
		for(j=0;j<info->flavor_info.flavor_info_len;j++){
			p = xdr_decode_sec_oid4(p,&rpcinfo->oid);
			rpcinfo->qop = ntohl(*p++);
			rpcinfo->service = ntohl(*p++);
			nfsv4_printk(secinfo,"NFSV4: xdr_decode_secinfo4 qop  %d  service %d\n", 
				rpcinfo->qop,rpcinfo->service);
			/* sec_oid4_len	4
			* sec_oid4-val, set to p, 4.	
			* qop	4
			* service	4
			* so move rpcinfo 16 char's
			*/
			(char *)rpcinfo += 16; 
		}
	}
	break;
	default:
	break;
	}
	return p;
}

#ifndef __OpenBSD__
/*
 * Encode void argument
 */
int
nfs4_xdr_enc_void(struct rpc_rqst *req, u32 *p, void *dummy)
{
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;
}
#endif

/*
 * Encode compound argument
 */
int
nfs4_xdr_compoundargs(struct rpc_rqst * req, u32 * p, struct COMPOUND4args * cargs)
{
    int             i;

    nfsv4_printk(compound, "NFSV4 RPC compoundargs encode START p: 0x%x opcnt: %d\n", (unsigned int) p, cargs->OPCNT);

    p = xdr_encode_utf8string(p, &cargs->tag);
    *p++ = htonl(cargs->minorversion);
    *p++ = htonl(cargs->OPCNT);
    for (i = 0; i < cargs->OPCNT; i++) {
	nfsv4_printk(compound, " opcode: %d \n", cargs->OPVAL[i].argop);
	*p++ = htonl(cargs->OPVAL[i].argop);
	switch (cargs->OPVAL[i].argop) {
	case OP_ACCESS:
	    {
		ACCESS4args    *argp = &cargs->OPVAL[i].OPU.opaccess;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_ACCESS. access: 0x%x\n", argp->access);
		*p++ = htonl(argp->access);
	    }
	    break;
	case OP_CLOSE:
	    {
		CLOSE4args     *argp = &cargs->OPVAL[i].OPU.opclose;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_CLOSE.\n");

		*p++ = htonl(argp->seqid);
		p = xdr_encode_64(p, argp->stateid);
	    }
	    break;
	case OP_CREATE:
	    {
		CREATE4args    *argp = &cargs->OPVAL[i].OPU.opcreate;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_CREATE.\n");

		p = xdr_encode_utf8string(p, &argp->objname);
		p = xdr_encode_createtype4(p, &argp->objtype);
	    }
	    break;
	case OP_GETATTR:
	    {
		GETATTR4args   *argp = &cargs->OPVAL[i].OPU.opgetattr;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_GETATTR\n");
		p = xdr_encode_bitmap4(p, &argp->attr_request);
	    }
	    break;
	case OP_GETFH:
	    nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_GETFH\n");
	    break;
	case OP_LINK:
	    {
		LINK4args      *argp = &cargs->OPVAL[i].OPU.oplink;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_LINK\n");
		p = xdr_encode_utf8string(p, &argp->newname);
	    }
	    break;
	case OP_LOOKUP:
	    {
		LOOKUP4args    *argp = &cargs->OPVAL[i].OPU.oplookup;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_LOOKUP\n");
		p = xdr_encode_pathname4(p, &argp->path);
	    }
	    break;
	case OP_LOOKUPP:
	    nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_LOOKUPP\n");
	    break;
	case OP_OPEN:
	    {
		OPEN4args      *argp = &cargs->OPVAL[i].OPU.opopen;

		nfsv4_printk(open, "NFSV4 RPC compoundargs - OP_OPEN\n");
		p = xdr_encode_open_claim4(p, &argp->claim);
		p = xdr_encode_openflag4(p, &argp->openhow);
		p = xdr_encode_nfs_lockowner4(p, &argp->owner);
		*p++ = htonl(argp->seqid);
		*p++ = htonl(argp->share_access);
		*p++ = htonl(argp->share_deny);
		nfsv4_printk(compound, "NFSV4 OPEN encode seqid %d, access %d, deny %d\n", argp->seqid, argp->share_access, argp->share_deny);
		nfsv4_printk(compound, "NFSV4 OPEN encode done\n");
	    }
	    break;
	case OP_OPEN_CONFIRM:
	    {
		OPEN_CONFIRM4args *argp = &cargs->OPVAL[i].OPU.opopen_confirm;

		nfsv4_printk(open,
		       "NFSV4 RPC encode compoundargs - OP_OPEN_CONFIRM\n");
		*p++ = htonl(argp->seqid);
		p = xdr_encode_verifier4(p, argp->open_confirm);
	    }
	    break;
	case OP_PUTFH:
	    {
		PUTFH4args     *argp = &cargs->OPVAL[i].OPU.opputfh;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_PUTFH\n");
		if (!(p = xdr_encode_fh4(p, &argp->object)))
		    return 0;
	    }
	    break;
	case OP_PUTROOTFH:
	    nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_PUTROOTFH\n");
	    break;
	case OP_READ:
	    {
		READ4args      *args = &cargs->OPVAL[i].OPU.opread;
		int             i;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_READ args->nr %d\n", args->nr);
		p = xdr_encode_64(p, args->stateid);
		p = xdr_encode_64(p, args->offset);
		*p++ = htonl(args->count);

#ifdef __linux__
		/*
		 * set up reply iovec. see nfs4_set_bsize for arg->iov set up
		 * for now, hard code  args-> nr = 2 or 3 depending if there
		 * are more compound calls after the READ *and* assuming
		 * there is no more than one OP_READ per compound call.
		 */
		req->rq_rvec[0].iov_len = args->iov[0].iov_len;
		req->rq_rvec[1].iov_len = args->iov[1].iov_len;
		req->rq_rvec[1].iov_base = args->iov[1].iov_base;
		nfsv4_printk(read, "AHHHHH rq_rvec[1].iov_base 0x%x args->iov[i].iov_base 0x%x\n",
			     (unsigned int) req->rq_rvec[0].iov_base,
			     (unsigned int) args->iov[1].iov_base);

		if (args->nr == 2) {
		    req->rq_rlen = args->count + args->iov[0].iov_len + args->iov[2].iov_len;
		    req->rq_rnr += 1;
		}
		if (args->nr == 3) {
		    req->rq_rvec[2].iov_len = args->iov[2].iov_len;
		    req->rq_rvec[2].iov_base = (u8 *) req->rq_rvec[0].iov_base + args->iov[0].iov_len;
		    req->rq_rlen = args->count + args->iov[0].iov_len + args->iov[2].iov_len;
		    req->rq_rnr += 2;
		}
		nfsv4_printk(read, "NFSV4 RPC compoundargs - req iov stuff. req->rq_rlen: %d req->rq_rnr %d\n", req->rq_rlen, req->rq_rnr);
		for (i = 0; i < req->rq_rnr; i++) {
		    nfsv4_printk(read, "req->rq_rvec[%d].iov_len %d\n", i, req->rq_rvec[i].iov_len);
		    nfsv4_printk(read, "BABY req->rq_rvec[%d].iov_base 0x%x\n", i, (unsigned int) req->rq_rvec[i].iov_base);
		}
#endif
	    }
	    break;
	case OP_READDIR:
	    {
		READDIR4args   *argp = &cargs->OPVAL[i].OPU.opreaddir;
		int             buflen = 0;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_READDIR p %p\n", p);
		p = xdr_encode_64(p, argp->cookie);
		p = xdr_encode_verifier4(p, argp->cookieverf);
		*p++ = htonl(argp->dircount);
		*p++ = htonl(argp->maxcount);
		p = xdr_encode_bitmap4(p, &argp->attr_request);

#ifdef __linux__
		/*
		 * set up reply iovec. see nfs4_set_bsize for arg->iov set
		 * up. for now, hard code  args-> nr = 3. assuming the call
		 * is in the form PUTFH READDIR GETATTR
		 */

		/*
		 * save the req->rq_rvec[0].iov_len which equals the bsize
		 * calculated by nfs4_set_bsize, which for each OP is the
		 * larger of args/ret. this size is used by rpc_allocate to
		 * kmalloc the rpc buffer used for both send and receive.
		 * req->rq_rvec[0].iov_base points to this buffer.
		 */
		buflen = req->rq_rvec[0].iov_len;
		nfsv4_printk(readdir, "NFSV4 READDIR save req->rq_rvec[0].iov_len %d iov_base %p\n", req->rq_rvec[0].iov_len, req->rq_rvec[0].iov_base);

		/*
		 * set the first iovec size to hold the PUTFH and READDIR
		 * args minus the READDIR data. the iov_base is already set
		 * into the rpc_alloc'ed buffer.
		 */
		req->rq_rvec[0].iov_len = argp->iov[0].iov_len;
		nfsv4_printk(readdir,
			     "READDIR IOVEC rq_rvec[0].iov_base 0x%x rq_rvec[0].iov_len %d\n",
			     (unsigned int) req->rq_rvec[0].iov_base,
			     (unsigned int) req->rq_rvec[0].iov_len);

		/* set the second iovec to the page cache page */
		req->rq_rvec[1].iov_len = argp->iov[1].iov_len;
		req->rq_rvec[1].iov_base = argp->iov[1].iov_base;
		nfsv4_printk(readdir,
			     "READDIR IOVEC rq_rvec[1].iov_base 0x%x rq_rvec[1].iov_len %d\n",
			     (unsigned int) req->rq_rvec[1].iov_base,
			     (unsigned int) req->rq_rvec[1].iov_len);

		/*
		 * set the third iovec to hold the end arguments of READDIR
		 * (i.e. the eof u_int) plus the GETATTR return. the iovbase
		 * will point into the rpc_alloc'ed buf past the first iovec
		 * data.
		 */
		req->rq_rvec[2].iov_len = buflen - argp->iov[0].iov_len;
		req->rq_rvec[2].iov_base = (u8 *) req->rq_rvec[0].iov_base + argp->iov[0].iov_len;
		nfsv4_printk(readdir,
			     "READDIR IOVEC rq_rvec[2].iov_base 0x%x rq_rvec[2].iov_len %d\n",
			     (unsigned int) req->rq_rvec[2].iov_base,
			     (unsigned int) req->rq_rvec[2].iov_len);

		req->rq_rlen = req->rq_rvec[0].iov_len + req->rq_rvec[1].iov_len + req->rq_rvec[2].iov_len;
		req->rq_rnr += 2;
		nfsv4_printk(readdir,
			   "READDIR IOVEC req->rq_rlen %d req->rq_rnr %d\n",
			     req->rq_rlen, req->rq_rnr);
#endif
	    }
	    break;
	case OP_READLINK:
	    /* eventually might want to use an iovec here, as in the other OP_READ * operations */
	    nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_READLINK\n");
	    break;
	case OP_REMOVE:
	    {
		REMOVE4args    *argp = &cargs->OPVAL[i].OPU.opremove;
		nfsv4_printk(compound, "NFSV4 RPC encode compoundargs - OP_REMOVE\n");
		p = xdr_encode_utf8string(p, &argp->target);
	    }
	    break;
	case OP_RENAME:
	    {
		RENAME4args    *argp = &cargs->OPVAL[i].OPU.oprename;
		nfsv4_printk(rename, "NFSV4 RPC encode compoundargs - OP_RENAME\n");
		p = xdr_encode_utf8string(p, &argp->oldname);
		p = xdr_encode_utf8string(p, &argp->newname);
		nfsv4_printk(rename,
			     "ERIC: old length %d, new length %d\n", argp->OLDNAME_LEN, argp->NEWNAME_LEN);
		nfsv4_printk(rename, "NFSV4 RPC encode OP_RENAME done\n");
	    }
	    break;
	case OP_RESTOREFH:
	    nfsv4_printk(rename, "NFSV4 RPC encode OP_RESTOREFH, nothing to do\n");
	    break;
	case OP_SAVEFH:
	    nfsv4_printk(rename, "NFSV4 RPC encode OP_SAVEFH, nothing to do\n");
	    break;
	case OP_SECINFO:
	    p = xdr_encode_utf8string(p, &cargs->OPVAL[i].OPU.opsecinfo.name);
	    break;
	case OP_SETCLIENTID:
	    {
		SETCLIENTID4args *argp = &cargs->OPVAL[i].OPU.opsetclientid;

		nfsv4_printk(setclntid,
			"NFSV4 RPC encode compoundargs - OP_SETCLIENTID\n");
		p = xdr_encode_nfs_client_id4(p, &argp->client);
		p = xdr_encode_cb_client4(p, &argp->callback);
	    }
	    break;
	case OP_SETCLIENTID_CONFIRM:
	    {
		SETCLIENTID_CONFIRM4args *argp = &cargs->OPVAL[i].OPU.opsetclientid_confirm;
		nfsv4_printk(setclntid,
		"NFSV4 RPC encode compoundargs - OP_SETCLIENTID_CONFIRM\n");
		p = xdr_encode_verifier4(p, argp->setclientid_confirm);
	    }
	    break;
	case OP_SETATTR:
	    {
		SETATTR4args   *argp = &cargs->OPVAL[i].OPU.opsetattr;

		nfsv4_printk(compound, "NFSV4 RPC compoundargs - OP_SETATTR\n");
		p = xdr_encode_64(p, argp->stateid);
		p = xdr_encode_fattr4(p, &argp->obj_attributes);
	    }
	    break;
	case OP_WRITE:
	    {
		WRITE4args     *argp = &cargs->OPVAL[i].OPU.opwrite;

		nfsv4_printk(compound, "NFSV4 encode compoundargs OP_WRITE:WRITELEN %d\n",
			     argp->WRITELEN);
		p = xdr_encode_64(p, argp->stateid);
		p = xdr_encode_64(p, argp->offset);
		*p++ = htonl(argp->stable);
		p = xdr_encode_write4data(p, argp);
	    }
	    break;
	default:
	    printk("NFSV4: encode: bad opcode\n");
	    /*
	     * NOTE: this is the return of rpc_call() in nfs4_proc_compound which
	     * returns a positive error from the wire. so, return positive errors
	     * in the xdr layer as well.
	     */
	    return EINVAL;
	    break;
	}
    }
    req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);

    nfsv4_printk(compound, "NFSV4 RPC compoundargs encode FINISH. p: %p req->rq_svec->iov_base %p req->rq_slen : %d req->rq_rlen %d \n", p, req->rq_svec->iov_base, req->rq_slen, req->rq_rlen);
    return 0;
}
/*
 * NFS XDR decode functions
 */
/*
 * Decode void reply
 */
int
nfs4_xdr_dec_void(struct rpc_rqst *req, u32 *p, void *dummy)
{
	return 0;
}

/*
 * Decode compound reply
 */

int
nfs4_xdr_compoundres(struct rpc_rqst *req, u32 *p, struct COMPOUND4res *cres)
{
    int i;

    nfsv4_printk(compound,"NFSV4 RPC compoundres decode START. p 0x%x\n",(unsigned int)p);
#ifdef CITI
    if ((*p & 0xffff0000) == 0xffff0000)
	cres->status = -(*p++);
    else
#endif
    cres->status = ntohl(*p++);

    nfsv4_printk(compound,"NFSV4 RPC compoundres decode cres->status %d Tag:  ",cres->status);

    p = xdr_decode_utf8string(p,&cres->tag);

    print_utf8string(&cres->tag);

    nfsv4_printk(compound,"NFSV4 RPC compoundres decode 1 p 0x%x\n",(unsigned int)p);
    cres->RESCNT = ntohl(*p++);

    nfsv4_printk(compound,"  rescnt %d\n",cres->RESCNT);

    for(i=0;i<cres->RESCNT;i++){
	cres->RESVAL[i].resop = ntohl(*p++);
	nfsv4_printk(compound," i: %d opcode: %d\n",i,cres->RESVAL[i].resop);
	switch(cres->RESVAL[i].resop){ 
	case OP_ACCESS:
	{
	    ACCESS4res *resp = &cres->RESVAL[i].RESU.opaccess;

	    nfsv4_printk(compound,"NFSV4 compound decode ACCESS status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);

	    if(resp->status == NFS4_OK){
		resp->SUPPORTED = ntohl(*p++);
		resp->ACCESS = ntohl(*p++);
	    }
	}
	    break;
	case OP_CLOSE:
	{
	    CLOSE4res *resp = &cres->RESVAL[i].RESU.opclose;

	    nfsv4_printk(compound,"NFSV4 compound decode CLOSE status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);

	    if(resp->status == NFS4_OK){
		p = xdr_decode_64(p,&resp->CLOSE4res_u.stateid);
		nfsv4_printk(compound,"NFSV4 compound decode CLOSE stateid:[0x%lx , %08lx]\n",
			     (unsigned long)(resp->CLOSE4res_u.stateid >> 32),
			     (unsigned long)(resp->CLOSE4res_u.stateid & 0xFFFFFFFFUL));
	    }
	}
	    break;
	case OP_CREATE:
	{
	    CREATE4res *resp = &cres->RESVAL[i].RESU.opcreate;

	    nfsv4_printk(compound, "NFSV4 compound decode CREATE status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	    if (resp->status == NFS4_OK){
		p = xdr_decode_change_info4(p,&resp->CREATEinfo);
		nfsv4_printk(compound, "CREATE atomic %d before %Ld after %Ld\n", resp->CREATEinfo.atomic, resp->CREATEinfo.before, resp->CREATEinfo.after);
	    }
	}
	    break;
	case OP_GETATTR:
	{
	    GETATTR4res *resp = &cres->RESVAL[i].RESU.opgetattr;

	    nfsv4_printk(compound,"NFSV4 compound decode GETTATTR status: "); 
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	    if(resp->status == NFS4_OK)
		p = xdr_decode_fattr4(p,&resp->GFATTR4);
	}
	    break;
	case OP_GETFH:
	{
	    GETFH4res *resp = &cres->RESVAL[i].RESU.opgetfh;

	    nfsv4_printk(compound,"NFSV4 compound decode GETFH status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	    if(resp->status == NFS4_OK)
		p = xdr_decode_fh4(p,&resp->GETFH4FH);
	}
	    break;
	case OP_LINK:
	{
	    LINK4res *resp = &cres->RESVAL[i].RESU.oplink;

	    nfsv4_printk(compound, "NFSV4 compound decode LINK status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound, "%d\n", resp->status);
	    resp->LINK_RES_INFO.atomic = ntohl(*p++);
	    nfsv4_printk(compound, "NFSDV4: LINK: decode RESULTS: atomic %d\n", resp->LINK_RES_INFO.atomic);
	    p = xdr_decode_64(p, &resp->LINK_RES_INFO.before);
	    p = xdr_decode_64(p, &resp->LINK_RES_INFO.after);
	    nfsv4_printk(compound, "NFSDV4: LINK: decode RESULTS: before %ld, after %ld\n", (long int)resp->LINK_RES_INFO.before, (long int)resp->LINK_RES_INFO.after);

	}
	    break;
	case OP_LOOKUP:
	{
	    LOOKUP4res *resp = &cres->RESVAL[i].RESU.oplookup;

	    nfsv4_printk(compound,"NFSV4 compound decode LOOKUP status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	}
	    break;
	case OP_LOOKUPP:
	{
	    LOOKUPP4res *resp = &cres->RESVAL[i].RESU.oplookupp;
	    resp->status = ntohl(*p++);
	    nfsv4_printk (compound, "NFSV4 compound decode LOOKUPP status: %d\n", resp->status);
	}
	    break;
	case OP_OPEN:
	{
	    OPEN4res *resp = &cres->RESVAL[i].RESU.opopen;
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound, "NFSV4 compound decode OPEN. status %d \n",resp->status);
	    if(resp->status == NFS4_OK){
		p = xdr_decode_64(p, &resp->OPEN_RES.stateid);
		p = xdr_decode_change_info4(p,&resp->OPEN_RES.cinfo);
		resp->OPEN_RES.rflags = ntohl(*p++);
		nfsv4_printk(open, "NFSV4 decode OPEN got rflags %d\n", resp->OPEN_RES.rflags);
		p = xdr_decode_verifier4(p,resp->OPEN_RES.open_confirm);
		p = xdr_decode_open_delegation4(p,&resp->OPEN_RES.delegation);
	    }
	}
	    break;
	case OP_OPEN_CONFIRM:
	{
	    OPEN_CONFIRM4res *resp = &cres->RESVAL[i].RESU.opopen_confirm;

	    nfsv4_printk(open,"NFSV4 compound decode OPEN_CONFIRM status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(open,"%d\n", resp->status);

	    if(resp->status == NFS4_OK){
				p = xdr_decode_64(p,&resp->OPEN_CON_RES.stateid);
				nfsv4_printk(open,"NFSV4 compound decode OPEN_CONFIRM stateid:[0x%lx , %08lx]\n",
			     (unsigned long)(resp->OPEN_CON_RES.stateid >> 32),
			     (unsigned long)(resp->OPEN_CON_RES.stateid & 0xFFFFFFFFUL));
	    }
	}
	    break;
	case OP_PUTFH:
	{
	    PUTFH4res *resp = &cres->RESVAL[i].RESU.opputfh;
	    nfsv4_printk(compound,"NFSV4 compound decode PUTFH p 0x%x status: ",(unsigned int)p);
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	}
	    break;
	case OP_PUTROOTFH:
	{
	    PUTROOTFH4res *resp = &cres->RESVAL[i].RESU.opputrootfh;
	
	    nfsv4_printk(compound,"NFSV4 compound decode PUTROOTFH status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	}
	    break;
	case OP_READ:
	{
	    READ4res *resp = &cres->RESVAL[i].RESU.opread;

	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"NFSV4 compound decode READ p %p status: %d\n", p, resp->status);
#ifndef __OpenBSD__
	    nfsv4_printk(read,"NFSV4 compound decode READ req->rq_rnr %d len 0 %d, len1 %d\n", req->rq_rnr,req->rq_rvec[0].iov_len,req->rq_rvec[1].iov_len);
#endif

	    if (resp->status == NFS4_OK) {
		resp->READEOF= ntohl(*p++);
		nfsv4_printk(compound,"EOF %d p %p ", resp->READEOF, p);
#ifdef __OpenBSD__
		req->rq_rvec->iov_base = p;
#endif
		p = xdr_decode_read4(p,resp);
		nfsv4_printk(compound,"READLEN %d ENDp %p\n", resp->READLEN, p);
	    }
	}
	    break;
	case OP_READDIR:
	{
	    READDIR4res *resp = &cres->RESVAL[i].RESU.opreaddir;
	    u32 *ptr;

	    nfsv4_printk(compound,"NFSV4 compound decode READDIR p %p status: ",p);
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	    if(resp->status == NFS4_OK){
		ptr = p;
#ifdef __OpenBSD__
		/*
		 * We don't actually decode anything here, it's done in the nfs_readdirrpc
		 * vnode op.  Note this means that p won't get updated to skip past the
		 * readdir response.  This is ok because it's the last op in the compound rpc.
		 * nfs_readdirrpc() uses the iov to find the dir entries.
		 */
		req->rq_rvec->iov_base = p;
#else
		p = xdr_decode_verifier4(p,resp->READDIR4res_u.resok4.cookieverf);
		if(!(p = xdr_decode_dirlist4(req,p,&resp->DIR4))){
		    printk("NFSV4 READDIR decode returning EINVAL\n");
			/* NOTE: this is the return of rpc_call() in nfs4_proc_compound 
			* which returns a positive error from the wire. so, return
			* positive errors in the xdr layer as well. 
			*/
		    return EINVAL; /*NOTE: check this return. should only be 
				     * RPC encoding errors*/
		}
#endif
	    }
	}
	    break;
	case OP_READLINK:
	{
	    READLINK4res *resp = &cres->RESVAL[i].RESU.opreadlink;
	    resp->status = ntohl (*p++);
	    if (resp->status == NFS4_OK)
		p = xdr_decode_utf8string (p, &resp->READLINK4res_u.resok4.link);
	}
	    break;
	case OP_REMOVE:
	{
	    REMOVE4res *resp = &cres->RESVAL[i].RESU.opremove;

	    nfsv4_printk(compound, "NFSV4 compound decode REMOVE status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound, "%d\n", resp->status);

	    if (resp->status == NFS4_OK){
		resp->REMOVE4res_u.resok4.cinfo.atomic = ntohl(*p++);
		nfsv4_printk(compound, "NFSV4: decode: REMOVE: atomic %d\n", resp->REMOVE4res_u.resok4.cinfo.atomic);
		p = xdr_decode_64(p, &resp->REMOVE4res_u.resok4.cinfo.before);
		nfsv4_printk(compound, "NFSV4: decode: REMOVE: before %Ld\n",  resp->REMOVE4res_u.resok4.cinfo.before);
		p = xdr_decode_64(p, &resp->REMOVE4res_u.resok4.cinfo.after);
		nfsv4_printk(compound, "NFSV4: decode: REMOVE: after %Ld\n", resp->REMOVE4res_u.resok4.cinfo.after);
	    }
	}
	    break;
	case OP_RENAME:
	{
	    RENAME4res *resp = &cres->RESVAL[i].RESU.oprename;
	    nfsv4_printk(rename, "NFSV4 compound decode RENAME status: ");

	    resp->status = ntohl(*p++);
	    nfsv4_printk(rename, "%d\n", resp->status);
	    if (resp->status == NFS4_OK) {
		resp->RENAME_RES_SRC.atomic = ntohl(*p++);
		nfsv4_printk(rename, "NFSDV4: RENAME: decode RESULTS: source atomic %d\n",
			     resp->RENAME_RES_SRC.atomic);
		p = xdr_decode_64(p, &resp->RENAME_RES_SRC.before);
		p = xdr_decode_64(p, &resp->RENAME_RES_SRC.after);
		nfsv4_printk(rename, "NFSDV4: RENAME: decode RESULTS: source before %ld, after %ld\n",
			     (long int)resp->RENAME_RES_SRC.before, (long int)resp->RENAME_RES_SRC.after);
		resp->RENAME_RES_TAR.atomic = ntohl(*p++);
		nfsv4_printk(rename, "NFSDV4: RENAME: decode RESULTS: target atomic %d\n", resp->RENAME_RES_TAR.atomic);
		p = xdr_decode_64(p, &resp->RENAME_RES_TAR.before);
		p = xdr_decode_64(p, &resp->RENAME_RES_TAR.after);
		nfsv4_printk(rename, "NFSDV4: RENAME: decode RESULTS: target before %ld, after %ld\n",
			     (long int)resp->RENAME_RES_TAR.before, (long int)resp->RENAME_RES_TAR.after);
		nfsv4_printk(rename, "NFSV4 compound decode RENAME done\n");
	    }
	}
	    break;
	case OP_RESTOREFH:
	{
	    RESTOREFH4res *resp = &cres->RESVAL[i].RESU.oprestorefh;
	    resp->status = ntohl(*p++);
	    nfsv4_printk (compound, "NFSV4 compound decode RESTOREFH status: %d\n", resp->status);
	}
	    break;
	case OP_SECINFO:
	{
	    SECINFO4res *resp = &cres->RESVAL[i].RESU.opsecinfo;
	    secinfo4 *sif4;
	    int i;

	    nfsv4_printk(secinfo,"NFSV4: compound decode SECINFO\n");
	    resp->status = ntohl(*p++);
	    switch(resp->status){
	    case NFS4_OK:
		nfsv4_printk(secinfo,"NFSV4: compound decode SECINFO NFS4_OK\n");
		resp->SECRES.SECINFO4resok_len = ntohl(*p++);	
		resp->SECRES.SECINFO4resok_val = kmalloc(resp->SECRES.SECINFO4resok_len, GFP_KERNEL);

		memset(resp->SECRES.SECINFO4resok_val,0,resp->SECRES.SECINFO4resok_len);
		sif4 = resp->SECRES.SECINFO4resok_val;
		nfsv4_printk(secinfo,
			     "NFSV4: compound decode SECINFO resok_len %d resok_val 0x%x \n",
			     resp->SECRES.SECINFO4resok_len,(unsigned int) resp->SECRES.SECINFO4resok_val);
		for(i=0;i<resp->SECRES.SECINFO4resok_len;i++)
		    p = xdr_decode_secinfo4(p,&resp->SECRES.SECINFO4resok_val[i]);
		break;
	    default:
		break;
	    }
	}
	    break;
	case OP_SETATTR:
	{
	    SETATTR4res *resp = &cres->RESVAL[i].RESU.opsetattr;

	    nfsv4_printk(compound, "NFSV4 compound decode SETATTR status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound, "%d\n", resp->status);
	    if (resp->status == NFS4_OK){
		p = xdr_decode_bitmap4(p, &resp->attrsset);
	    }
	}
	    break;
	case OP_SETCLIENTID:
	{
	    SETCLIENTID4res *resp = &cres->RESVAL[i].RESU.opsetclientid;

	    nfsv4_printk(setclntid,"NFSV4: compound decode SETCLIENTID\n");
	    resp->status = ntohl(*p++);
	    switch(resp->status){
	    case NFS4_OK:
		p = xdr_decode_64(p,&resp->SETCLIENTID4res_u.resok4.clientid);
		p = xdr_decode_verifier4(p,resp->SETCLIENTID4res_u.resok4.setclientid_confirm);
		break;
	    case NFS4ERR_CLID_INUSE:
		p = xdr_decode_clientaddr4(p,&resp->SETCLIENTID4res_u.client_using);
		break;
	    default:
		break;
	    }
	}
	    break;
	case OP_SETCLIENTID_CONFIRM:
	{
	    SETCLIENTID_CONFIRM4res *resp = &cres->RESVAL[i].RESU.opsetclientid_confirm;

	    nfsv4_printk(setclntid,"NFSV4: compound decode SETCLIENTID_CONFIRM\n");
	    resp->status = ntohl(*p++);
	}
	    break;
	case OP_SAVEFH:
	{
	    SAVEFH4res *resp = &cres->RESVAL[i].RESU.opsavefh;
	    nfsv4_printk(rename, "NFSV4 compound decode SAVEFH p 0x%x status: ",(unsigned int)p);

	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
	}      
	    break;
	case OP_WRITE:
	{
	    WRITE4res *resp = &cres->RESVAL[i].RESU.opwrite;

	    nfsv4_printk(compound,"NFSV4 compound decode WRITE status: ");
	    resp->status = ntohl(*p++);
	    nfsv4_printk(compound,"%d\n", resp->status);
#ifdef __OpenBSD__
	    req->rq_rvec->iov_base = p;
#else
	    if(resp->status == NFS4_OK){
		p = xdr_decode_write4(p,resp);
	    }
#endif
	}
	    break;
	default:
	    printk("NFSV4: decode: bad opcode\n");
			/* NOTE: this is the return of rpc_call() in nfs4_proc_compound 
			* which returns a positive error from the wire. so, return
			* positive errors in the xdr layer as well. 
			*/
	    return EINVAL;
	    break;
	}
    }
    nfsv4_printk(compound,"NFSV4 RPC compound decode FINISH status %d\n",cres->status);
    if (cres->status < 0)
	cres->status = -(cres->status);
    if((cres->status < 0) || (cres->status > NFS4ERR_LEASE_MOVED)){
	printk("NFSV4 compound decode ILLEGAL STATUS %d - set to inval\n",cres->status);
	cres->status = NFS4ERR_INVAL;
    }
    return(cres->status);
}


/*
 * Calculate maximum space in bytes needed for a fattr4 reply 
 * e.g. use NFS4_MAXNAMLEN for utf8strings.
 */
int
get_fattr4_res_buf_size(bitmap4 *bm4)
{
int i,j;
int size=0;
int bit;

	if(!bm4)
		return 0;
	/* for each bit set in the bit mask */
	for(i=0;i<bm4->BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4->BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				nfsv4_printk(attributes,"NFSV4:get_fattr4_res_buf_size bit %d is set\n",bit);
				/* calculate size for correct return argument*/
				switch(bit){
				case FATTR4_SUPPORTED_ATTRS:
					size += sizeof(u_int); /* bitmap4 len */
					/* don't really know how many attributes
					* a server will support. right now the
					* maximum number of attributes supported 
					* by the protocol is 54 which fits into 
					* two uint32_t. FATTR4_ATTR_MAX is not in 
					* spec  - see nfs4_prot.h
					*/
					size += sizeof(u_int) * (FATTR4_ATTR_MAX +1)/32;
				break;
				case FATTR4_TYPE:
					size += sizeof(fattr4_type);
				break;
				case FATTR4_FH_EXPIRE_TYPE:
					size += sizeof(fattr4_fh_expire_type);
				break;
				case FATTR4_CHANGE:
					size += sizeof(fattr4_change);
				break;
				case FATTR4_SIZE:
					size += sizeof(fattr4_size);
				break;
				case FATTR4_LINK_SUPPORT:
					size += sizeof(fattr4_link_support);
				break;
				case FATTR4_SYMLINK_SUPPORT:
					size += sizeof(fattr4_symlink_support);
				break;
				case FATTR4_NAMED_ATTR:
					size += sizeof(fattr4_named_attr);
				break;
				case FATTR4_FSID:
					size += sizeof(fattr4_fsid);
				break;
				case FATTR4_UNIQUE_HANDLES:
					size += sizeof(fattr4_unique_handles);
				break;
				case FATTR4_LEASE_TIME:
					size += sizeof(fattr4_lease_time);
				break;
				case FATTR4_RDATTR_ERROR:
					size += sizeof(fattr4_rdattr_error);
				break;
				case FATTR4_ACL:
					size += sizeof(fattr4_acl);
				break;
				case FATTR4_ACLSUPPORT:
					size += sizeof(fattr4_aclsupport);
				break;
				case FATTR4_ARCHIVE:
					size += sizeof(fattr4_archive);
				break;
				case FATTR4_CANSETTIME:
					size += sizeof(fattr4_cansettime);
				break;
				case FATTR4_CASE_INSENSITIVE:
					size += sizeof(fattr4_case_insensitive);
				break;
				case FATTR4_CASE_PRESERVING:
					size += sizeof(fattr4_case_preserving);
				break;
				case FATTR4_CHOWN_RESTRICTED:
					size += sizeof(fattr4_chown_restricted);
				break;
				case FATTR4_FILEHANDLE:
					size += sizeof(u_int); /* nfs_fh4_len */
				size += NFS4_FHSIZE;
				break;
				case FATTR4_FILEID:
					size += sizeof(fattr4_fileid);
				break;
				case FATTR4_FILES_AVAIL:
					size += sizeof(fattr4_files_avail);
				break;
				case FATTR4_FILES_FREE:
					size += sizeof(fattr4_files_free);
				break;
				case FATTR4_FILES_TOTAL:
					size += sizeof(fattr4_files_total);
				break;
				case FATTR4_HIDDEN:
					size += sizeof(fattr4_hidden);
				break;
				case FATTR4_HOMOGENEOUS:
					size += sizeof(fattr4_homogeneous);
				break;
				case FATTR4_MAXFILESIZE:
					size += sizeof(fattr4_maxfilesize);
				break;
				case FATTR4_MAXLINK:
					size += sizeof(fattr4_maxlink);
				break;
				case FATTR4_MAXNAME:
					size += sizeof(fattr4_maxname);
				break;
				case FATTR4_MAXREAD:
					size += sizeof(fattr4_maxread);
				break;
				case FATTR4_MAXWRITE:
					size += sizeof(fattr4_maxwrite);
				break;
				case FATTR4_MIMETYPE:
					size += sizeof(u_int); 
				size += NFS4_MAXNAMLEN;
				break;
				case FATTR4_MODE:
					size += sizeof(fattr4_mode);
				break;
				case FATTR4_NO_TRUNC:
					size += sizeof(fattr4_no_trunc);
				break;
				case FATTR4_NUMLINKS:
					size += sizeof(fattr4_numlinks);
				break;
				case FATTR4_OWNER:
					size += sizeof(u_int); 
				size += NFS4_MAXNAMLEN;
				break;
					case FATTR4_OWNER_GROUP:
					size += sizeof(u_int); 
					size += NFS4_MAXNAMLEN;
				break;
				case FATTR4_QUOTA_HARD:
					size += sizeof(fattr4_quota_hard);
				break;
				case FATTR4_QUOTA_SOFT:
					size += sizeof(fattr4_quota_soft);
				break;
				case FATTR4_QUOTA_USED:
					size += sizeof(fattr4_quota_used);
				break;
				case FATTR4_RAWDEV:
					size += sizeof(fattr4_rawdev);
				break;
				case FATTR4_SPACE_AVAIL:
					size += sizeof(fattr4_space_avail);
				break;
				case FATTR4_SPACE_FREE:
					size += sizeof(fattr4_space_free);
				break;
				case FATTR4_SPACE_TOTAL:
					size += sizeof(fattr4_space_total);
				break;
				case FATTR4_SPACE_USED:
					size += sizeof(fattr4_space_used);
				break;
				case FATTR4_SYSTEM:
					size += sizeof(fattr4_system);
				break;
				case FATTR4_TIME_ACCESS:
					size += sizeof(fattr4_time_access);
				break;
				case FATTR4_TIME_ACCESS_SET:
					size += sizeof(fattr4_time_access_set);
				break;
				case FATTR4_TIME_BACKUP:
					size += sizeof(fattr4_time_backup);
				break;
				case FATTR4_TIME_CREATE:
					size += sizeof(fattr4_time_create);
				break;
				case FATTR4_TIME_DELTA:
					size += sizeof(fattr4_time_delta);
				break;
				case FATTR4_TIME_METADATA:
					size += sizeof(fattr4_time_metadata);
				break;
				case FATTR4_TIME_MODIFY:
					size += sizeof(fattr4_time_modify);
				break;
				case FATTR4_TIME_MODIFY_SET:
					size += sizeof(fattr4_time_modify_set);
				break;
				default:
				break;
				}
			}
		}
	}
	nfsv4_printk(attributes,"NFSV4 get_fattr4_res_buf_size: #bytes: %d\n",size);
	return size;
}

/* return # xdr words in pathname4 */
int
get_pathname4_bufsize(pathname4 *p4)
{
int j,size = 0;

	size += 1; /* pathname4_len */
	for(j=0;j<p4->pathname4_len;j++){

		size += 1; /* utfstring_len*/

		size += XDR_QUADLEN(p4->pathname4_val[j].utf8string_len);

	}
	return size;
}

/* dynamically set the compound rpc p_bufsize, in # of 32bit words. 
 * follow nfsv2,v3 and use the maximum of send, recv args.
 * bsize = size of send and recv buf for xdr'ed data
 *
 * NOTE: for now, assume that PUTFH READ is the ONLY 
 * 	compound call made by this client that contains a READ.
 * return_size = size of decoded return data for a PUTFH,READ call, 
 * 		minus the actual READ data size plus the
 * 		rpc header size.
 * XXX Right now only recvsize is returned, and it's the max of recv and send sizes.
 */
int
nfs4_get_sizes(u32 *sendsizep, u32 *recvsizep, struct COMPOUND4args *args, struct COMPOUND4res *res)
{
    int i;
    u_int bsize = 0;
    u_int ret_size = 0;		/* for READ iovec */

    nfsv4_printk(bsize,"nfs4_set_bsize BEGINNING bsize %d \n",bsize);
    bsize += 1;			/* top level status */
    bsize += 1;			/* tag utf_stringlen variable */


    /* xdr tag string length */
    bsize += XDR_QUADLEN(args->tag.utf8string_len); 

    nfsv4_printk(bsize,"nfs4_set_bsize:tag len %d XDR_QUADLEN %d\n",
		 args->tag.utf8string_len,
		 XDR_QUADLEN(args->tag.utf8string_len));
	
    bsize += 1;			/* data_len*/

    ret_size = bsize;
    nfsv4_printk(bsize,"nfs4_set_bsize before OPCNT loop  bsize %d \n",bsize);
    for(i=0;i<args->OPCNT;i++){
	ret_size += 1;		/* argop,resop */
	bsize += 1;		/* data_len*/

	switch(args->OPVAL[i].argop){ 
	case OP_ACCESS:
	    /* res larger than args*/
	    bsize += 3;		/*status,access,supported*/
	    nfsv4_printk(bsize,"ACCESS nfs4_set_bsize bsize %d \n",bsize);
	    break;
	case OP_CLOSE:
	    /* res, args, same size*/
	    bsize += 3;		/*status,stateid*/
	    nfsv4_printk(bsize,"CLOSE nfs4_set_bsize bsize %d \n",bsize);
	    break;
	case OP_CREATE:
	{
	    CREATE4args *argp = &args->OPVAL[i].OPU.opcreate;
	    int asize = 0, rsize = 0;

	    asize += XDR_QUADLEN(argp->objname.utf8string_len);
	    asize += 1;		/* nfs_ftype4 */
	    switch (argp->objtype.type){
	    case (NF4LNK):
		asize += XDR_QUADLEN(argp->objtype.createtype4_u.linkdata.utf8string_len);
		break;
	    case (NF4BLK):
	    case (NF4CHR):
		asize += 1;	/* specdata1 */
		asize += 1;	/* specdata2 */
		break;
	    default:
		break;
	    }
	    rsize += 1;		/* atomic */
	    rsize += 2;		/* before */
	    rsize += 2;		/* after */

	    if (asize >= rsize) bsize += asize;
	    else bsize += rsize;	
	    nfsv4_printk(bsize,"CREATE nfs4_set_bsize bsize %d \n",bsize);
	}
	    break;
	case OP_GETATTR:
	{
	    GETATTR4args *argp = &args->OPVAL[i].OPU.opgetattr;
	    GETATTR4res *resp = &res->RESVAL[i].RESU.opgetattr;

	    /* NOTE: to be cool, since i already called get_fattr4_res_buf_size
	     * to kmalloc the result structures, before calling nfs4_proc_compound
	     * i should set the res attr_len to the
	     * result of the get_fattr4_res_buf_size, use it to increment bsize, and 
	     * then zero the attr_len, thus avoid a call to get_fattr_res_buf_size()
	     * in nfs4_set_bsize().
	     * e.g.
	     * bsize += res->RESVAL[i].RESU.opgetattr.GFATTR4.FATTRSLEN
	     * res->RESVAL[i].RESU.opgetattr.GFATTR4.FATTRSLEN = 0;
	     */
	    /* res larger than args*/

	    /* result size =
	     *    1                                       (status)
	     *  + 1                                       (bitmap len)
	     *  + argp->attr_request.BLEN                 (bitmap val)
	     *  + 1                                       (attlist len)
	     *  + XDR_QUADLEN(resp->GFATTR4.FATTRSLEN)    (attrlist val)
	     *  = 3 + argp->attr_request.BLEN + XDR_QUADLEN(resp->GFATTR4.FATTRSLEN)
	     */
	    int result_size = 3 + argp->attr_request.BLEN + XDR_QUADLEN(resp->GFATTR4.FATTRSLEN);

	    bsize += result_size;
	    ret_size += result_size;
	    resp->GFATTR4.FATTRSLEN = 0;
		
	    nfsv4_printk(bsize,"GETATTR nfs4_set_bsize bsize %d \n",bsize);
	}
	    break; 
	case OP_GETFH:
	    /* res larger than args */
	    bsize += 1;		/*status*/
	    bsize += 1;		/*nfs_fh4_len*/
	    bsize += (NFS4_FHSIZE>>2); /* max file handle size */
	    nfsv4_printk(bsize,"GETFH nfs4_set_bsize bsize %d \n",bsize);
	    break;
	case OP_LINK:
	{
	    LINK4args *argp = &args->OPVAL[i].OPU.oplink;
	    nfsv4_printk(compound, "NFSV4: setting bsize for OP_LINK\n");
	    if (XDR_QUADLEN(argp->newname.utf8string_len) >= 5){
		/* args bigger than results */
		bsize += 1;	/* newname length */
		bsize += XDR_QUADLEN(argp->newname.utf8string_len);
	    }else{
		/* results bigger than args */
		bsize += 1;	/* status */
		bsize += 1;	/* atomic */
		bsize += 2;	/* before */
		bsize += 2;	/* after */
	    }
	    nfsv4_printk(bsize,"LINK nfs4_set_bsize bsize %d \n",bsize);
	    nfsv4_printk(compound, "NFSV4: bszie LINK done\n");
	}
	    break;
	case OP_LOOKUP:
	{
	    bsize += get_pathname4_bufsize(&args->OPVAL[i].OPU.oplookup.path);
	    nfsv4_printk(bsize,"LOOKUP nfs4_set_bsize bsize %d \n",bsize);
	}
	    break;
	case OP_LOOKUPP:
	{
	    ret_size += 1;
	    bsize += 1;		/* status */
	    nfsv4_printk (bsize, "LOOKUPP nfs4_set_bsize bsize %d\n", bsize);
	}
	    break;
	case OP_OPEN:
	{
	    /*args larger than ret (?)*/
	    OPEN4args *argp = &args->OPVAL[i].OPU.opopen;
	    open_claim4 *c4 = &argp->claim;
	    openflag4 	*of4 =  &argp->openhow;
	    createhow4 	*ch4 = &of4->openflag4_u.how;
	    nfs_lockowner4 *lown = &argp->owner;
	    open_claim_delegate_cur4 *cur4 = &c4->open_claim4_u.delegate_cur_info;

	    nfsv4_printk(bsize,"OPEN nfs4_set_bsize beginning bsize %d \n",bsize);

	    /* claim */
	    switch(c4->claim){
	    case CLAIM_NULL:
		bsize += get_pathname4_bufsize(&c4->open_claim4_u.file);
		break;
	    case CLAIM_PREVIOUS:
		bsize += 1;	/* delegate_type */
		break;
	    case CLAIM_DELEGATE_CUR:
		bsize += get_pathname4_bufsize(&cur4->file);
		bsize += 2;	/* delegate_stateid */
		break;
	    case CLAIM_DELEGATE_PREV:
		bsize += get_pathname4_bufsize(&c4->open_claim4_u.file_delegate_prev);
		break;
	    default:
		nfsv4_printk(xdr,"NFSV4 nfs4_set_bsize: invlaid open_claim4 switch %d\n",c4->claim);
		break;
	    }

	    /* openhow */
	    bsize += 1;		/*opentype*/
	    switch (of4->opentype){
	    case OPEN4_CREATE:
	    {
		bsize += 1; 	/*mode*/
		switch(ch4->mode){
		case UNCHECKED4:
		case GUARDED4:
		    /* bsize - words, get_fattr4_res_buf_size() - bytes */
		    bsize += 1; /* bitmap4_len */
		    bsize += argp->openhow.HOW.CREATE_ATTR.attrmask.BLEN; 
		    bsize += 1; /* attrlist len */
		    bsize += XDR_QUADLEN(argp->openhow.HOW.CREATE_ATTR.FATTRSLEN);
		    break;
		case EXCLUSIVE4:
		    bsize += XDR_QUADLEN(NFS4_VERIFIER_SIZE);
		    break;
		default:
		    nfsv4_printk(xdr,"NFSV4 nfs4_set_bsizw: invlaid createhow4 mode %d\n",ch4->mode);
		    break;
		}
	    }
		break;
	    case OPEN4_NOCREATE:
		break;
	    default:
		nfsv4_printk(xdr,"NFSV4 nfs4_set_bsize: invlaid openflag4 switch %d\n",of4->opentype);
		break;
	    }

	    /* nfs_lockowner4 */
	    bsize += 2;		/* clientid */
	    bsize += XDR_QUADLEN(lown->owner.owner_len);

	    /* seqid */
	    bsize += 1; 
	    /* share_access*/
	    bsize += 1; 
	    /* share_deny */
	    bsize += 1; 
	    nfsv4_printk(bsize,"OPEN nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;
	case OP_PUTFH:
	    /* args larger than return */

	    bsize += 1;
	    bsize += XDR_QUADLEN(NFS4_FHSIZE);

	    ret_size += 1;	/*status*/
	    nfsv4_printk(bsize,"PUTFH nfs4_set_bsize end bsize %d \n",bsize);

	    break;
	case OP_PUTROOTFH:
	    bsize += 1;		/*status*/
	    nfsv4_printk(bsize,"PUTROOTFH nfs4_set_bsize end bsize %d \n",bsize);

	    break;
	case OP_READ:
	{
	    READ4args *argp= &args->OPVAL[i].OPU.opread;
		
	    /* args larger than res because using seperate
	     * iovec for returned data 
	     */
	    bsize += 2;		/*stateid*/
	    bsize += 2;		/*offset*/
	    bsize += 1;		/*count*/
	    /* set size in bytes of return iovec up to now. 
	     * iov_base is set in sunrpc/clnt.c:call_encode()  
	     */ 
	    ret_size += 1;	/* status*/
	    ret_size += 1;	/* eof */
	    ret_size += 1;	/* read_len */
	    /* clnt.c: call_verify - uses 6 words of rpc header.
	     * should this be RPC_REPHDRSIZE + req->rq_task.tk_auth->au_slack 
	     * instead of 6 ?
	     */
	    ret_size += 6;

	    argp->iov[0].iov_len = ret_size << 2;
	    /* set size of readdir iovec */
	    argp->iov[1].iov_len = argp->count;
	    argp->nr = 2;
			
	    nfsv4_printk(read,"nfs4_set_bsize: iov[0].iov_len %d, iov[1].iov_len %d, bsize %d\n",
			 argp->iov[0].iov_len,
			 argp->iov[1].iov_len,
			 bsize);
	    nfsv4_printk(bsize,"READ nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;
	case OP_READDIR:
	{
	    READDIR4args *argp = &args->OPVAL[i].OPU.opreaddir;
		
	    bsize += 1;		/*status*/
	    bsize += XDR_QUADLEN(NFS4_VERIFIER_SIZE); /*verifier4*/
	    bsize += 1;		/* eof */
	    nfsv4_printk(bsize,"READDIR nfs4_set_bsize end bsize %d \n",bsize);

	    ret_size += 1;	/* status */
	    ret_size += XDR_QUADLEN(NFS4_VERIFIER_SIZE); /*verifier4*/
	    /* clnt.c: call_verify - uses 6 words of rpc header.
	     * should this be RPC_REPHDRSIZE + req->rq_task.tk_auth->au_slack 
	     * instead of 6 ?
	     */
	    ret_size += 6;
	    argp->iov[0].iov_len = ret_size << 2;
	    nfsv4_printk(kendrick, "SET_BSIZE iov[0].iov_len %d\n",argp->iov[0].iov_len);
	    /* set size of read iovec */
			
	    argp->iov[1].iov_len = argp->maxcount;
	    argp->nr = 2;
			
	    nfsv4_printk(readdir, "nfs4_set_bsize: readdir iov[0].iov_len %d, iov[1].iov_len %d, bsize %d\n",
			 argp->iov[0].iov_len,
			 argp->iov[1].iov_len,
			 bsize);
	}
	    break;
	case OP_READLINK:
	    bsize += 1  // status
		  +  1  // link len
		  +  XDR_QUADLEN (NFS4_MAXNAMLEN); // link val (have to allocate maximum possible)
	    break;
	case OP_REMOVE:
	{
	    REMOVE4args *argp = &args->OPVAL[i].OPU.opremove;

	    if (XDR_QUADLEN(argp->target.utf8string_len) > 6){
		bsize +=1;	/* target length */
		bsize += XDR_QUADLEN(argp->target.utf8string_len); /* target name */
	    }else{
		/* status 1
		 * atomic 1
		 * before 2
		 * after 2
		 */
		bsize += 6;
	    }
	    nfsv4_printk(bsize, "REMOVE nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;	
	case OP_RENAME:
	{
	    RENAME4args *argp = &args->OPVAL[i].OPU.oprename;
	    nfsv4_printk(rename, "NFSV4: RENAME: setting bsize\n");
	    if ( (XDR_QUADLEN(argp->OLDNAME_LEN) ) + (XDR_QUADLEN(argp->NEWNAME_LEN) ) >=9){
		/* args are bigger than results */
		bsize += 1;	/* oldname length */
		bsize += XDR_QUADLEN(argp->OLDNAME_LEN); /* oldname val */
		bsize += 1;	/* newname length */
		bsize += XDR_QUADLEN(argp->NEWNAME_LEN); /* newname val */
	    }else{
		/* results are bigger than args : total = 11 words
		 * status 1
		 * source.atomic 1
		 * source.before 2
		 * source.after 2
		 * target.atomic 1
		 * target.before 2
		 * target.after 2
		 */
		bsize += 11;
	    }
	    nfsv4_printk(bsize,"RENAME nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;
	case OP_RESTOREFH:
	{
	    bsize += 1;            // return status
	    ret_size += 1;         // return status
	}
	    break;
	case OP_SAVEFH:
	{
	    bsize += 1;            // return status
	    ret_size += 1;         // return status
	}
	    break;
	case OP_SETATTR:           // arg is larger than return val
	{
	    SETATTR4args *argp = &args->OPVAL[i].OPU.opsetattr;
	    nfsv4_printk(secinfo, "NFSV4: SETATTR: setting bsize\n");

	    bsize += 2;		/* stateid */
	    bsize += 1;		/* bitmap len */
	    bsize += argp->obj_attributes.attrmask.bitmap4_len; /* bitmap vals */

	    bsize += 1;		/* attrlist len */
	    bsize += XDR_QUADLEN(argp->obj_attributes.attr_vals.attrlist4_len);
		
	    nfsv4_printk(bsize,"SETATTR nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break; 
	case OP_SECINFO:
	{
	    /* this is really stupid. we need to stop guessing how big the returned
	     * rpc size is, and calculate the actual size on the server.
	     * we don't know what if anything will be returned, so give enough space
	     * for a couple of rpcsec_gss_info structs .....
	     */
	    nfsv4_printk(secinfo, "NFSV4: SECINFO: setting bsize\n");

	    bsize += 1;		/*status*/
	    bsize += 1;		/* SECINFO4resok_len */

	    /* rpcsec_gss_info
	     * sec_oid4_len 	1
	     * sec_oid4_val  64   (a complete guess)
	     * qop4	1
	     * service 1
	     * so, 67, per rpcsec_gss_info plus 1 for the flavor_len = 68
	     * 68 * 2 = 136
	     */
	    bsize += 136;

	    nfsv4_printk(bsize,"SECINFO nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;
	case OP_SETCLIENTID:
	{
	    SETCLIENTID4args *argp = &args->OPVAL[i].OPU.opsetclientid;
	    /* args larger than response.
	     * clientid	2
	     * cb_client4.cb_program 1
	     * cb_client4.cb_program 1
	     */ 
	    bsize += 4;
	    bsize += strlen(argp->callback.cb_location.r_netid);
	    bsize += strlen(argp->callback.cb_location.r_addr);
	    nfsv4_printk(bsize,
			 "SETCLIENTID nfs4_set_bsize end bsize %d \n",bsize);
	}
	    break;
	case OP_SETCLIENTID_CONFIRM:
	    /*args larger than return*/
	    bsize += NFS4_VERIFIER_SIZE;
	    nfsv4_printk(bsize,
			 "SETCLIENTID_CONFIRM nfs4_set_bsize end bsize %d \n",bsize);
	    break;
	default:
	    break;
	} /*switch*/
    }

    if (bsize > 1024)
	return NFS4ERR_RESOURCE;

    /* Size in bytes is quadwords times 4 */
    *sendsizep = bsize << 2;
    *recvsizep = ret_size << 2;

    return 0;
}

#ifdef __linux__
int
nfs4_set_bsize(struct rpc_clnt *clnt, u32 proc, struct COMPOUND4args *args, struct COMPOUND4res *res)
{
    u32 sendsize, recvsize;
    int err;

    if ((err = nfs4_get_sizes(&sendsize, &recvsize, args, res)))
	return err;

    clnt->cl_procinfo[proc].p_bufsiz = recvsize;

    nfsv4_printk(bsize,"NFSV4 nfs4_set_bsize size: %d\n",clnt->cl_procinfo[proc].p_bufsiz);
    return 0;
}

#ifndef MAX
# define MAX(a, b)	(((a) > (b))? (a) : (b))
#endif

#define PROC(proc, argtype, restype)	\
	{ "nfs_" #proc,					\
	(kxdrproc_t) nfs4_xdr_##argtype,			\
	(kxdrproc_t) nfs4_xdr_##restype,			\
	MAX(NFS_##argtype##_sz,NFS_##restype##_sz) << 2	\
}

static struct rpc_procinfo	nfs4_procedures[2] = {
  PROC(null,	 enc_void,	dec_void),
  PROC(compound, compoundargs, compoundres),
};

struct rpc_version		nfs_version4 = {
	4,
	sizeof(nfs4_procedures)/sizeof(nfs4_procedures[0]),
	nfs4_procedures
};
#endif
